import re
import click
from tabulate import tabulate
import logging
from typing import List, Optional
from requests.exceptions import HTTPError
import functools

from datetime import datetime, timedelta
from tzlocal import get_localzone

from togglcmder.toggl.cli.workspaces \
    import sync_or_retrieve_workspaces, retrieve_workspace_from_context

from togglcmder.toggl.cli.projects \
    import sync_or_retrieve_projects, retrieve_project_from_context

from togglcmder.toggl.cli.tags import sync_or_retrieve_tags

from togglcmder.toggl.cli.helpers import retrieve_cache_from_context
from togglcmder.toggl.cli.helpers import retrieve_commands_from_context
from togglcmder.toggl.cli.helpers import retrieve_downloader_from_context

from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.filters.workspaces import Workspaces as WorkspaceFilter

from togglcmder.toggl.types.project import Project
from togglcmder.toggl.filters.projects import Projects as ProjectFilter

from togglcmder.toggl.types.tag import Tag
from togglcmder.toggl.filters.tags import Tags as TagFilter

from togglcmder.toggl.types.time_entry import TimeEntry
from togglcmder.toggl.builders.time_entry_builder import TimeEntryBuilder
from togglcmder.toggl.views.time_entry import TimeEntry as TimeEntryView
from togglcmder.toggl.filters.time_entries import TimeEntries as TimeEntryFilter


def validated_time(date_time: str) -> datetime:
    result = re.match(r'^now(?P<value>[+-]\d+)?(?P<unit>[ydhms])?$',
                      date_time)
    if result is not None:
        now = get_localzone().localize(datetime.now())
        value = result.group('value')
        if not value:
            return now
        if result.group('unit') == 'd':
            return now + timedelta(days=float(value))
        elif result.group('unit') == 'h':
            return now + timedelta(hours=float(value))
        elif result.group('unit') == 'm':
            return now + timedelta(minutes=float(value))
        elif result.group('unit') == 's':
            return now + timedelta(seconds=float(value))

    try:
        return get_localzone().localize(datetime.fromisoformat(date_time))
    except ValueError:
        raise RuntimeError(f"Not a valid time: '{date_time}'.")


def sync_or_retrieve_time_entries(context_obj: dict, workspace: Workspace, start: datetime, stop: datetime, *,
                                  project: Optional[Project] = None) -> List[TimeEntry]:
    caching = retrieve_cache_from_context(context_obj)
    downloader = retrieve_downloader_from_context(context_obj)

    # Download any new time entries if remote sync is enabled.
    if context_obj['sync'] is True:
        current_time_entries = downloader.download_time_entries(start, stop)
        if current_time_entries:
            caching.update_time_entry_cache(current_time_entries)
    else:
        # Otherwise we just pull from the local cache.
        current_time_entries = caching.retrieve_time_entry_cache()

    # Neither downloading from remote nor pulling from local cache does any filtering, so let's do that.
    current_time_entries = TimeEntryFilter.filter_on_workspace(current_time_entries, workspace)

    # It also doesn't filter on start/stop, so let's do that too.
    current_time_entries = TimeEntryFilter.filter_on_date_range(current_time_entries, start, stop)

    # Time entries _can_ be filtered on project if it is provided.
    current_time_entries = TimeEntryFilter.filter_on_project(current_time_entries, project)

    return current_time_entries


@click.group(
    help='Add, update, delete, start, stop, and list timers.'
)
@click.option('--workspace')
@click.option('--project')
@click.option('--download-start',
              type=validated_time,
              help='This can be now[-/+[dhms]] or an iso formatted time string.')
@click.option('--download-stop',
              type=validated_time,
              help='This can be now[-/+[dhms]] or an iso formatted time string.')
@click.pass_context
def timers(context: click.Context,
           workspace: str,
           project: str,
           download_start: datetime,
           download_stop: datetime):
    workspace_name = workspace or context.obj['config']['default_workspace']
    project_name = project or context.obj['config']['default_project']

    workspaces = sync_or_retrieve_workspaces(context.obj)
    if workspace_name:
        # If a workspace name is provided, default or otherwise, filter on it.
        workspaces = WorkspaceFilter.filter_on_name(
            workspaces,
            workspace_name)

        # If a single workspace was found we can store it.
        workspaces_found = len(workspaces)
        if workspaces_found == 1:
            context.obj['data']['workspace'] = workspaces[0]

    context.obj['data']['workspaces'] = workspaces

    projects = []
    # Whether or not we have workspaces filtered, we can download all projects.
    for current_workspace in workspaces:
        synced_projects = sync_or_retrieve_projects(context.obj, current_workspace)
        if synced_projects:
            projects.extend(synced_projects)

    if project_name and not projects:
        # No projects exist in this workspace, but the user requested to filter on one.
        return

    if projects and project_name:
        # If there is a specified project name then we can filter on it.
        projects = ProjectFilter.filter_on_name(
            projects,
            project_name)

        # If a single project was found, we can store it too.
        projects_found = len(projects)
        if projects_found == 1:
            context.obj['data']['project'] = projects[0]
    context.obj['data']['projects'] = projects

    tags = []
    for current_workspace in workspaces:
        synced_tags = sync_or_retrieve_tags(context.obj, current_workspace)
        if synced_tags:
            tags.extend(synced_tags)
    context.obj['data']['tags'] = tags

    # Use our default window settings to download time entry updates.
    if not download_start:
        then = get_localzone().localize(
            datetime.now() - timedelta(days=context.obj['config']['default_time_entry_window_start_days']))
    else:
        then = download_start

    if not download_stop:
        now = get_localzone().localize(
            datetime.now() - timedelta(days=context.obj['config']['default_time_entry_window_stop_days']))
    else:
        now = download_stop

    time_entries = []
    for current_workspace in workspaces:
        for current_project in projects:
            if current_project.workspace_identifier == current_workspace.identifier:
                synced_entries = sync_or_retrieve_time_entries(
                    context.obj, current_workspace, then, now, project=current_project)
                if synced_entries:
                    time_entries.extend(synced_entries)
        if not projects:
            synced_entries = sync_or_retrieve_time_entries(
                context.obj, current_workspace, then, now)
            if synced_entries:
                time_entries.extend(synced_entries)

    context.obj['data']['time_entries'] = time_entries


@timers.command(
    'add',
    short_help='Add a completed time entry.',
    help='Add a new completed time entry. Stop time OR duration must be provided to add a valid '
         'completed entry.')
@click.option('--description')
@click.option('--start-time',
              type=validated_time,
              help='This can be now[-/+[dhms]] or an iso formatted time string.',
              required=True)
@click.option('--stop-time',
              type=validated_time,
              help='This can be now[-/+[dhms]] or an iso formatted time string.')
@click.option('--duration',
              help='The duration of this time entry. Stop time will be calculated if this is provided.',
              type=int)
@click.option('--tags')
@click.pass_context
def timer_add(context: click.Context,
              description: str,
              start_time: datetime,
              stop_time: datetime,
              duration: int,
              tags: str):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when adding a new timer.')
        return

    # Handling the error situations:
    if not stop_time and not duration:
        click.echo(click.style('ERROR', fg='red')
                   + ': you must specify a stop time or duration when adding a new timer.')
        return

    time_entries = context.obj['data']['time_entries']
    time_entries = TimeEntryFilter.filter_on_description(
        time_entries, description
    )

    if time_entries and description:
        click.echo(click.style('ERROR', fg='red')
                   + ': a timer with this description already exists.')
        return

    time_entry_builder = TimeEntryBuilder()
    time_entry_builder.workspace_identifier(workspace.identifier)

    project = retrieve_project_from_context(context)
    if project:
        time_entry_builder.project_identifier(project.identifier)

    if description:
        time_entry_builder.description(description)
    if start_time:
        time_entry_builder.start_time(dt=start_time)
    if stop_time:
        time_entry_builder.stop_time(dt=stop_time)
    if duration:
        time_entry_builder.duration(duration)
    elif stop_time:
        time_entry_builder.duration(
            int(stop_time.timestamp() - start_time.timestamp()))
    if tags:
        time_entry_builder.tags(tags.split(','))

    caching = retrieve_cache_from_context(context.obj)
    commands = retrieve_commands_from_context(context.obj)

    try:
        time_entry = time_entry_builder.build()
        time_entry = commands.add_completed_time_entry(time_entry)
        caching.update_time_entry_cache([time_entry])

        click.echo(click.style('SUCCESS', fg='green') +
                   f': added new time entry to'
                   f' workspace({workspace.name})!')

    except HTTPError as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to add new time entry to workspace({workspace.name}).'
                   ' An exception has been logged; check the logs for more information.')


@timers.command('delete')
@click.option('--description')
@click.option('--multiple',
              is_flag=True,
              default=False,
              show_default=True)
@click.option('--tags')
@click.pass_context
def timer_delete(context: click.Context, description: str, multiple: bool, tags: str):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when deleting a time entry.')
        return

    if not description and not tags:
        click.echo(click.style('ERROR', fg='red') +
                   f': either description or tags must be specified when deleting a timer.')
        return

    time_entries = context.obj['data']['time_entries']

    if description:
        time_entries = TimeEntryFilter.filter_on_description(
            time_entries, description
        )
        if not time_entries:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no time entries exist with the specified description.')
            return

    current_tags = context.obj['data']['tags']
    if tags:
        current_tags = TagFilter.filter_on_names(
            current_tags, tags.split(',')
        )
        if not current_tags:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no tags exist with the specified names.')
            return

        time_entries = TimeEntryFilter.filter_on_any_tags(
            time_entries, current_tags
        )
        if not time_entries:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no time entries exist with the specified tags.')
            return

    if len(time_entries) > 1 and not multiple:
        click.echo(click.style('ERROR', fg='red') +
                   ': multiple time entries found matching the criteria, but --multiple '
                   'not specified.. please use the --multiple option or tighten '
                   'the search criteria.')
        return

    caching = retrieve_cache_from_context(context.obj)
    commands = retrieve_commands_from_context(context.obj)

    try:
        for entry in time_entries:
            commands.delete_time_entry(entry)
            caching.remove_time_entry_from_cache(entry)
            click.echo(click.style('SUCCESS', fg='green') +
                       f': deleted timer({entry.description})!')

    except HTTPError as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to delete the timer(s). An exception has been logged;'
                   ' check the logs for more information.')


@timers.command('update',
                short_help='Update an existing time entry with new details.',
                help='Update an existing time entry with new details.')
@click.option('--old-description')
@click.option('--new-description')
@click.option('--new-project')
@click.option('--old-tags')
@click.option('--add-tags')
@click.option('--remove-tags')
@click.option('--new-start-time')
@click.option('--new-duration')
@click.option('--new-stop-time')
@click.option('--multiple',
              is_flag=True,
              default=False,
              show_default=True)
@click.pass_context
def timer_update(context: click.Context,
                 old_description: str,
                 new_description: str,
                 new_project: str,
                 old_tags: str,
                 add_tags: str,
                 remove_tags: str,
                 new_start_time: datetime,
                 new_duration: int,
                 new_stop_time: datetime,
                 multiple: bool):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when updating a timer.')
        return

    if not old_description or not old_tags:
        click.echo(click.style('ERROR', fg='red') +
                   f': either description or tags must be specified when deleting a timer.')
        return

    time_entries = context.obj['data']['time_entries']
    if not time_entries:
        click.echo(click.style('ERROR', fg='red') +
                   f': no time entries to update in this workspace({workspace.name}).')
        return

    if old_description:
        time_entries = TimeEntryFilter.filter_on_description(
            time_entries, old_description
        )
        if not time_entries:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no time entries exist with the specified description.')
            return

    current_tags = context.obj['data']['tags']
    if old_tags:
        current_tags = TagFilter.filter_on_names(
            current_tags, old_tags.split(',')
        )
        if not current_tags:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no tags exist with the specified names.')
            return

        time_entries = TimeEntryFilter.filter_on_any_tags(
            time_entries, current_tags
        )
        if not time_entries:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no time entries exist with the specified tags.')
            return
        # restore current_tags so we can filter it again if needed.
        current_tags = context.obj['data']['tags']

    if len(time_entries) > 1 and not multiple:
        click.echo(click.style('ERROR', fg='red') +
                   ': multiple timers found matching the criteria, but --multiple '
                   'not specified.. please used the --multiple option or tighten '
                   'the search criteria.')
        return

    projects = context.obj['data']['projects']
    new_specified_project: Optional[Project] = None
    if new_project:
        projects = ProjectFilter.filter_on_name(
            projects, new_project
        )
        if not projects:
            click.echo(click.style('ERROR', fg='red') +
                       f': no projects found with specified name.')
            return
        elif len(projects) > 1:
            click.echo(click.style('ERROR', fg='red') +
                       f': multiple projects found with specified name. Update criteria '
                       f' to match a single project.')
            return
        new_specified_project = projects[0]

    additional_tags: Optional[List[Tag]] = None
    if add_tags:
        current_tags = TagFilter.filter_on_names(
            current_tags, add_tags.split(',')
        )
        if not current_tags:
            click.echo(click.style('ERROR', fg='red') +
                       f': no tags found with the specified names in --add-tags.')
            return
        # restore current_tags so we can filter it again if needed.
        current_tags = context.obj['data']['tags']

    removed_tags: Optional[List[Tag]] = None
    if remove_tags:
        current_tags = TagFilter.filter_on_names(
            current_tags, remove_tags.split(',')
        )
        if not current_tags:
            click.echo(click.style('ERROR', fg='red') +
                       f': no tags found with the specified names in --remove-tags.')
            return

    entries: List[TimeEntry] = []
    for time_entry in time_entries:
        time_entry_builder = TimeEntryBuilder(time_entry)
        if new_description:
            time_entry_builder.description(new_description)
        if new_project:
            time_entry_builder.project_identifier(new_specified_project.identifier)
        if add_tags:
            time_entry_builder.add_tags([tag.name for tag in additional_tags])
        if remove_tags:
            time_entry_builder.remove_tags([tag.name for tag in removed_tags])
        if new_start_time:
            time_entry_builder.start_time(dt=new_start_time)
        if new_duration:
            time_entry_builder.duration(new_duration)
        if new_stop_time:
            time_entry_builder.stop_time(dt=new_stop_time)
        entries.append(time_entry_builder.build())

    caching = retrieve_cache_from_context(context.obj)
    commands = retrieve_commands_from_context(context.obj)

    try:
        updated_entries: List[TimeEntry] = []
        for entry in entries:
            updated_entry = commands.update_completed_time_entry(entry)
            updated_entries.append(updated_entry)
            click.echo(click.style('SUCCESS', fg='green') +
                       f': updated time entry({updated_entry.description}) in '
                       f' workspace({workspace.name})!')
        caching.update_time_entry_cache(updated_entries)

    except HTTPError as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to update the timer(s). An exception has been logged;'
                   ' check the logs for more information.')

@timers.command('start',
                short_help='Start a new running timer.',
                help='Start a new running timer. Has an implicit start time of '
                     'now. If no workspace or project is provided, defaults will '
                     'be used. If there are no defaults, an error will be issued.')
@click.option('--description')
@click.option('--tags')
@click.pass_context
def timer_start(context: click.Context, description: str, tags: str):
    time_entry_builder = TimeEntryBuilder()

    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when starting a timer.')
        return
    time_entry_builder.workspace_identifier(workspace.identifier)

    current_tags = context.obj['data']['tags']
    if tags:
        current_tags = TagFilter.filter_on_names(
            current_tags, tags.split(',')
        )
        if not current_tags:
            click.echo(click.style('ERROR', fg='red') +
                       f': no tags match the specified names.')
            return
        time_entry_builder.tags([tag.name for tag in current_tags])

    project = retrieve_project_from_context(context)
    if project:
        time_entry_builder.project_identifier(project.identifier)

    if description:
        time_entry_builder.description(description)

    time_entry_builder.start_time(dt=datetime.now(tz=get_localzone()))

    caching = retrieve_cache_from_context(context.obj)
    commands = retrieve_commands_from_context(context.obj)

    try:
        entry = time_entry_builder.build()
        entry = commands.start_time_entry(entry)
        click.echo(click.style('SUCCESS', fg='green') +
                   f': started new time entry({entry.description}) in'
                   f' workspace({workspace.name}).')
        caching.update_time_entry_cache([entry])

    except HTTPError as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to start the timer. An exception has been logged;'
                   ' check the logs for more information.')


@timers.command('stop',
                help='Stop the current running timer, if one exists.')
@click.pass_context
def timer_stop(context: click.Context):
    downloader = retrieve_downloader_from_context(context.obj)
    entry = downloader.get_current_time_entry()

    if not entry:
        click.echo('No entry is currently running.')
        return

    commands = retrieve_commands_from_context(context.obj)
    caching = retrieve_cache_from_context(context.obj)

    try:
        entry = commands.stop_time_entry(entry)
        caching.update_time_entry_cache([entry])
        click.echo(click.style('SUCCESS', fg='green') +
                   f': stopped the current running timer({entry.description}).')

    except HTTPError as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to stop the current running timer. An exception'
                   f' has been logged; check the logs for more information.')


@timers.command('current',
                help='Get the current running timer, if one exists.')
@click.pass_context
def timer_current(context: click.Context):
    project = retrieve_project_from_context(context)
    workspace = retrieve_workspace_from_context(context)

    downloader = retrieve_downloader_from_context(context.obj)

    try:
        entry = downloader.get_current_time_entry()
        if not entry:
            click.echo('No entry is currently running.')
        else:
            time_entry_view = TimeEntryView(
                [entry],
                project,
                workspace)
            click.echo_via_pager(tabulate(
                time_entry_view.values_sorted(),
                headers=TimeEntryView.headers(),
                tablefmt="grid"))

    except HTTPError as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to download the current running timer. An exception'
                   f' has been logged; check the logs for more information.')


@timers.command('resume',
                short_help='Resume the most recent timer.',
                help='Look through the list of timers, and resume the most recent one, '
                     'regardless of project or workspace.')
@click.pass_context
def timer_resume(context: click.Context):
    time_entries = context.obj['data']['time_entries']
    time_entries = sorted(time_entries, key=lambda entry: entry.last_updated,
                            reverse=True)

    commands = retrieve_commands_from_context(context.obj)
    caching = retrieve_cache_from_context(context.obj)

    try:
        time_entry_builder = TimeEntryBuilder(time_entries[0])
        time_entry_builder.start_time(dt=datetime.now(get_localzone()))
        time_entry_builder.unset_stop_time()
        time_entry_builder.unset_duration()
        started_entry = commands.start_time_entry(time_entry_builder.build())
        if started_entry:
            caching.update_time_entry_cache([started_entry])
        click.echo(click.style('SUCCESS', fg='green', bold=True) +
                   f': Resumed time entry({started_entry.description}).')
    except HTTPError as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to resume the latest timer. An exception'
                   f' has been logged; check the logs for more information.')


@timers.command('list')
@click.option('--description')
@click.option('--sort-by',
              type=click.Choice([val.__str__() for val in TimeEntryView.headers()]),
              default=TimeEntryView.headers()[5],
              show_default=True)
@click.pass_context
def timer_list(context: click.Context, description: str, sort_by: str):
    workspaces = context.obj['data']['workspaces']
    projects = context.obj['data']['projects']
    time_entries = context.obj['data']['time_entries']

    time_entries = TimeEntryFilter.filter_on_description(
        time_entries, description
    )
    if not time_entries:
        click.echo(click.style('WARNING', fg='yellow') +
                   f': no entries found in the specified workspace(s).')
        return

    rows = []
    for current_workspace in workspaces:
        reduced_entries = []
        # get all of the descriptions available
        descriptions = [entry.description for entry in time_entries]
        # remove any duplicates
        descriptions = list(dict.fromkeys(descriptions))
        for current_description in descriptions:
            # filter the actual entries on each unique description
            entries_on_description = TimeEntryFilter.filter_on_description(time_entries, current_description)
            # reduce down the entries with that description into a single entry
            combined_entry = functools.reduce(lambda x, y:
                                              TimeEntryBuilder(x).duration(x.duration + y.duration).build(),
                                              entries_on_description)
            reduced_entries.append(combined_entry)

        for entry in reduced_entries:
            project = ProjectFilter.filter_on_identifier(projects,
                                                         entry.project_identifier)
            rows.extend(
                TimeEntryView(
                    [entry], project[0] if project else None, current_workspace
                ).values()
            )

    # Sort by modified date by default.
    sort_by_index = 5
    if sort_by:
        if sort_by not in TimeEntryView.headers():
            click.echo(f'Choose a valid header to sort by: {TimeEntryView.headers()}')
            return
        for index, header in enumerate(TimeEntryView.headers()):
            if sort_by == header:
                sort_by_index = index
                break

    rows = sorted(rows, key=lambda key: key[sort_by_index], reverse=True)

    click.echo_via_pager(tabulate(
        rows,
        headers=TimeEntryView.headers(),
        tablefmt="grid"))



import unittest
import json
from datetime import datetime, timedelta
from pytz import timezone

from togglcmder.toggl.builders.time_entry_builder import TimeEntryBuilder
from togglcmder.toggl.builders.project_builder import ProjectBuilder
from togglcmder.toggl.builders.tag_builder import TagBuilder

from togglcmder.toggl.encoders.project_encoder import ProjectEncoder
from togglcmder.toggl.encoders.time_entry_encoder import TimeEntryEncoder
from togglcmder.toggl.encoders.tag_encoder import TagEncoder


class TestEncoders(unittest.TestCase):
    def test_project_encoding(self):
        expected_project = {
            'project': {
                'name': 'Project Name',
                'wid': 1234,
                'color': 1,
            }
        }

        project = ProjectBuilder()\
            .name(expected_project['project']['name'])\
            .workspace_identifier(expected_project['project']['wid'])\
            .color(expected_project['project']['color'])\
            .build()

        self.assertDictEqual(json.loads(json.dumps(project, cls=ProjectEncoder)),
                             expected_project)

    def test_tag_encoding(self):
        expected_tag = {
            'tag': {
                'name': 'Tag Name',
                'wid': 1234
            }
        }

        tag = TagBuilder()\
            .name(expected_tag['tag']['name'])\
            .workspace_identifier(expected_tag['tag']['wid'])\
            .build()

        self.assertDictEqual(json.loads(json.dumps(tag, cls=TagEncoder)), expected_tag)

    def testing_time_entry_encoding(self):
        current_time = datetime.now(timezone('EST')).replace(microsecond=0)
        delta_time = timedelta(hours=2)
        stop_time = current_time + delta_time
        expected_time_entry = {
            'time_entry': {
                'description': 'Some time entry.',
                'tags': [
                    'Some tag goes here.',
                    'Another tag goes here.'
                ],
                'pid': 1234,
                'wid': 1,
                'start': current_time.isoformat(),
                'duration': delta_time.seconds,
                'stop': stop_time.isoformat(),
                'created_with': 'toggl_cmder'
            }
        }

        time_entry = TimeEntryBuilder()\
            .description(expected_time_entry['time_entry']['description'])\
            .workspace_identifier(expected_time_entry['time_entry']['wid'])\
            .project_identifier(expected_time_entry['time_entry']['pid'])\
            .start_time(start_time=current_time.isoformat())\
            .duration(delta_time.seconds)\
            .stop_time(stop_time=stop_time.isoformat())\
            .tags(expected_time_entry['time_entry']['tags'])\
            .build()

        self.assertDictEqual(json.loads(json.dumps(time_entry, cls=TimeEntryEncoder)),
                             expected_time_entry)


if __name__ == '__main__':
    unittest.main()



from togglcmder.toggl.caching import Caching
from togglcmder.toggl.downloader import Downloader
from togglcmder.toggl.commands import Commands


# Helpers that don't fit anywhere else!
def retrieve_downloader_from_context(context: dict) -> Downloader:
    downloader = context['downloader']
    assert(isinstance(downloader, Downloader))
    return downloader


def retrieve_cache_from_context(context: dict) -> Caching:
    caching = context['cache']
    assert(isinstance(caching, Caching))
    return caching


def retrieve_commands_from_context(context: dict) -> Commands:
    commands = context['commands']
    assert(isinstance(commands, Commands))
    return commands



import unittest
import urllib.parse

from datetime import datetime
from pytz import timezone

from togglcmder.toggl.endpoints.api import API


class TestEndpoints(unittest.TestCase):
    ENDPOINT_URL_FORMAT = "{endpoint}/v{version}/{type}{arguments}"

    DEFAULT_ENDPOINT = "https://www.toggl.com/api"
    DEFAULT_VERSION = 8

    def test_time_entries_base(self):
        time_entry_url = API(version=TestEndpoints.DEFAULT_VERSION).time_entries.__repr__()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="time_entries",
            arguments=""
        )

        if time_entry_url != expected_url:
            self.fail("{} ne {}".format(expected_url, time_entry_url))

    def test_time_entries_details(self):
        time_entry_id = 1

        time_entry_url = API(version=TestEndpoints.DEFAULT_VERSION).time_entries.details(time_entry_id)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="time_entries",
            arguments="/{}".format(time_entry_id)
        )

        if time_entry_url != expected_url:
            self.fail("{} ne {}".format(expected_url, time_entry_url))

    def test_multiple_time_entries_details(self):
        time_entry_ids = [1, 2, 3, 4]

        time_entry_url = API(version=TestEndpoints.DEFAULT_VERSION).time_entries.details(identifiers=time_entry_ids)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="time_entries",
            arguments="/{}".format(",".join(map(str, time_entry_ids)))
        )

        if time_entry_url != expected_url:
            self.fail("{} ne {}".format(expected_url, time_entry_url))

    def test_time_entries_start(self):
        time_entry_url = API(version=TestEndpoints.DEFAULT_VERSION).time_entries.start()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="time_entries/start",
            arguments=""
        )

        if time_entry_url != expected_url:
            self.fail("{} ne {}".format(expected_url, time_entry_url))

    def test_time_entries_stop(self):
        time_entry_id = 1

        time_entry_url = API(version=TestEndpoints.DEFAULT_VERSION).time_entries.stop(time_entry_id)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="time_entries",
            arguments="/{}/stop".format(time_entry_id)
        )

        if time_entry_url != expected_url:
            self.fail("{} ne {}".format(expected_url, time_entry_url))

    def test_time_entries_search(self):
        # The dates used in the generated URLs"
        start_date = datetime(2019, 12, 1, 10, 0, 0, tzinfo=timezone('EST'))
        end_date = datetime(2019, 12, 1, 11, 0, 0, tzinfo=timezone('EST'))

        # The URL under test:
        time_entry_url = API(version=TestEndpoints.DEFAULT_VERSION).time_entries.search(
            start=start_date,
            end=end_date)

        # The expected result as generated here in the test case:
        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="time_entries",
            arguments="?start_date={start_date}&end_date={end_date}".format(
                start_date=urllib.parse.quote(start_date.isoformat()),
                end_date=urllib.parse.quote(end_date.isoformat())
            )
        )

        if time_entry_url != expected_url:
            self.fail("{} ne {}".format(expected_url, time_entry_url))

    def test_time_entries_current(self):
        time_entry_url = API(version=TestEndpoints.DEFAULT_VERSION).time_entries.current()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="time_entries/current",
            arguments=""
        )

        if time_entry_url != expected_url:
            self.fail("{} ne {}".format(expected_url, time_entry_url))

    def test_projects_base(self):
        projects_url = API(version=TestEndpoints.DEFAULT_VERSION).projects.__repr__()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="projects",
            arguments=""
        )

        if projects_url != expected_url:
            self.fail("{} ne {}".format(expected_url, projects_url))

    def test_projects_details(self):
        project_id = 1

        projects_url = API(version=TestEndpoints.DEFAULT_VERSION).projects.details(identifier=project_id)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="projects",
            arguments="/{}".format(project_id)
        )

        if projects_url != expected_url:
            self.fail("{} ne {}".format(expected_url, projects_url))

    def test_multiple_projects_details(self):
        project_ids = [1, 2, 3, 4]

        projects_url = API(version=TestEndpoints.DEFAULT_VERSION).projects.details(identifiers=project_ids)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="projects",
            arguments="/{}".format(",".join(map(str, project_ids)))
        )

        if projects_url != expected_url:
            self.fail("{} ne {}".format(expected_url, projects_url))

    def test_tags_base(self):
        tags_url = API(version=TestEndpoints.DEFAULT_VERSION).tags.__repr__()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="tags",
            arguments=""
        )

        if tags_url != expected_url:
            self.fail("{} ne {}".format(expected_url, tags_url))

    def test_tags_details(self):
        tag_id = 1

        tags_url = API(version=TestEndpoints.DEFAULT_VERSION).tags.details(tag_id)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="tags",
            arguments="/{}".format(tag_id)
        )

        if tags_url != expected_url:
            self.fail("{} ne {}".format(expected_url, tags_url))

    def test_users_base(self):
        users_url = API(version=TestEndpoints.DEFAULT_VERSION).users.__repr__()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="me",
            arguments=""
        )

        if users_url != expected_url:
            self.fail("{} ne {}".format(expected_url, users_url))

    def test_users_details(self):
        users_url = API(version=TestEndpoints.DEFAULT_VERSION).users.details()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="me",
            arguments=""
        )

        if users_url != expected_url:
            self.fail("{} ne {}".format(expected_url, users_url))

    def test_users_details_with_extra(self):
        users_url = API(version=TestEndpoints.DEFAULT_VERSION).users.details(extra_data=True)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="me",
            arguments="?with_related_data=true"
        )

        if users_url != expected_url:
            self.fail("{} ne {}".format(expected_url, users_url))

    def test_reset_api_token(self):
        users_url = API(version=TestEndpoints.DEFAULT_VERSION).users.reset_api_token()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="reset_token",
            arguments=""
        )

        if users_url != expected_url:
            self.fail("{} ne {}".format(expected_url, users_url))

    def test_workspaces_base(self):
        users_url = API(version=TestEndpoints.DEFAULT_VERSION).workspaces.__repr__()

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="workspaces",
            arguments=""
        )

        if users_url != expected_url:
            self.fail("{} ne {}".format(expected_url, users_url))

    def test_workspaces_details(self):
        workspace_id = 1

        users_url = API(version=TestEndpoints.DEFAULT_VERSION).workspaces.details(workspace_id)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="workspaces",
            arguments="/{}".format(workspace_id)
        )

        if users_url != expected_url:
            self.fail("{} ne {}".format(expected_url, users_url))

    def test_workspaces_projects(self):
        workspace_id = 1

        users_url = API(version=TestEndpoints.DEFAULT_VERSION).workspaces.projects(workspace_id)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="workspaces",
            arguments="/{}/projects".format(workspace_id)
        )

        if users_url != expected_url:
            self.fail("{} ne {}".format(expected_url, users_url))

    def test_workspaces_tags(self):
        workspace_id = 1

        users_url = API(version=TestEndpoints.DEFAULT_VERSION).workspaces.tags(workspace_id)

        expected_url = TestEndpoints.ENDPOINT_URL_FORMAT.format(
            endpoint=TestEndpoints.DEFAULT_ENDPOINT,
            version=TestEndpoints.DEFAULT_VERSION,
            type="workspaces",
            arguments="/{}/tags".format(workspace_id)
        )

        if users_url != expected_url:
            self.fail("{} ne {}".format(expected_url, users_url))


if __name__ == '__main__':
    unittest.main()



from togglcmder.toggl.endpoints.time_entries import TimeEntries
from togglcmder.toggl.endpoints.projects import Projects
from togglcmder.toggl.endpoints.tags import Tags
from togglcmder.toggl.endpoints.workspaces import Workspaces
from togglcmder.toggl.endpoints.users import Users


class API(object):
    def __init__(self, version: int = 8):
        self.__version = version
        self.__url = "https://www.toggl.com/api/v{}".format(
            self.__version
        )

    @property
    def time_entries(self) -> TimeEntries:
        return TimeEntries(url=self.__url)

    @property
    def projects(self) -> Projects:
        return Projects(url=self.__url)

    @property
    def tags(self) -> Tags:
        return Tags(url=self.__url)

    @property
    def workspaces(self) -> Workspaces:
        return Workspaces(url=self.__url)

    @property
    def users(self) -> Users:
        return Users(url=self.__url)



from typing import List, Tuple

from togglcmder.toggl.types.project import Project as ProjectType
from togglcmder.toggl.types.workspace import Workspace as WorkspaceType


class Project(object):
    """
    A composition of the data we want to show in a view
    of a given project.
    """

    def __init__(self, *, projects: List[ProjectType], workspace: WorkspaceType):
        self.__projects = projects
        self.__workspace = workspace

    @staticmethod
    def headers() -> Tuple:
        return (
            "Project Identifier",
            "Project Name",
            "Workspace Name",
            "Color Designation",
            "Last Updated"
        )

    def values(self) -> List[Tuple]:
        return [(
            project.identifier,
            project.name,
            self.__workspace.name,
            project.color,
            project.last_updated.isoformat()
        ) for project in self.__projects]



from enum import Enum, unique
from typing import Optional
from datetime import datetime
from functools import total_ordering


@total_ordering
class Project(object):
    @unique
    class Color(Enum):
        LIGHT_BLUE = 0
        LIGHT_PURPLE = 1
        PINK = 2
        LIGHT_ORANGE = 3
        ORANGE = 4
        LIGHT_GREEN = 5
        CYAN = 6
        PEACH = 7
        BLUE = 8
        PURPLE = 9
        YELLOW = 10
        GREEN = 11
        RED = 12
        BRIGHT_RED = 13
        BLACK = 14

        @staticmethod
        def from_string(string: str):
            return Project.Color[string.upper()]

        def __str__(self):
            return self.name.lower()

    def __init__(self, *, name: str, workspace_identifier: int,
                 color: Optional[Color] = None,
                 identifier: Optional[int] = None,
                 last_updated: Optional[datetime] = None,
                 created: Optional[datetime] = None):
        self.__identifier = identifier
        self.__name = name
        self.__workspace_identifier = workspace_identifier
        self.__color = color
        self.__last_updated = last_updated
        self.__created = created

    def __eq__(self, other):
        if not isinstance(other, Project):
            return NotImplemented

        return self.identifier == other.identifier \
            and self.name == other.name \
            and self.workspace_identifier == other.workspace_identifier \
            and self.color == other.color \
            and self.last_updated == other.last_updated \
            and self.created == other.created

    def __lt__(self, other):
        if not isinstance(other, Project):
            return NotImplemented
        return self.__name.lower() < other.name.lower()

    def __str__(self):
        return "{},{},{},{},{},{}".format(
            self.name,
            self.identifier,
            self.last_updated,
            self.workspace_identifier,
            self.color.name,
            self.created)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def identifier(self) -> int:
        return self.__identifier

    @property
    def workspace_identifier(self) -> int:
        return self.__workspace_identifier

    @property
    def color(self) -> Color:
        return self.__color

    @property
    def last_updated(self) -> datetime:
        return self.__last_updated

    @property
    def created(self) -> datetime:
        return self.__created



from json import JSONEncoder

from togglcmder.toggl.types.project import Project


class ProjectEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Project):
            project = {'project': {
                'name': obj.name,
                'wid': obj.workspace_identifier,
            }}
            if obj.color is not None:
                project['project']['color'] = obj.color.value
            return project
        return super(ProjectEncoder, self).default(obj)



import json

from togglcmder.toggl.types.time_entry import TimeEntry


class TimeEntryEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, TimeEntry):
            entry = {'time_entry': {
                'wid': obj.workspace_identifier,
                'start': obj.start_time.isoformat(),
                'tags': obj.tags,
                'created_with': 'toggl_cmder'
            }}
            if obj.project_identifier is not None:
                entry['time_entry']['pid'] = obj.project_identifier
            if obj.description is not None:
                entry['time_entry']['description'] = obj.description
            if obj.duration is not None and obj.duration > 0:
                entry['time_entry']['duration'] = obj.duration
            if obj.stop_time is not None:
                entry['time_entry']['stop'] = obj.stop_time.isoformat()
            return entry
        return super(TimeEntryEncoder, self).default(obj)



from typing import List
import re as regex
import logging

from togglcmder.toggl.types.workspace import Workspace


class Workspaces(object):

    @staticmethod
    def filter_single(
            workspaces: List[Workspace],
    ) -> Workspace:
        if not workspaces:
            raise RuntimeError('no workspaces passed in when single workspace expected')
        elif len(workspaces) > 1:
            raise RuntimeError('multiple workspaces passed in when single workspace expected')

        return workspaces[0]

    @staticmethod
    def filter_on_name(
            workspaces: List[Workspace],
            name: str
    ) -> List[Workspace]:
        logger = logging.getLogger(__name__)
        if not name:
            logger.debug('no name provided for filter_on_name(workspaces) with input = {}'.format(
                [workspace.name for workspace in workspaces]
            ))
            return workspaces

        result = list(filter(
            lambda workspace:
                regex.match("^{}$".format(name),
                            workspace.name,
                            regex.IGNORECASE),
            workspaces
        ))

        if not result:
            logger.warning('no workspaces found with name = {} and with input = {}'.format(
                name, [workspace.name for workspace in workspaces]
            ))
        return result

    @staticmethod
    def filter_on_identifier(
            workspaces: List[Workspace],
            identifier: int
    ) -> List[Workspace]:
        logger = logging.getLogger(__name__)
        if not identifier:
            logger.debug('no identifier provided for filter_on_identifier(workspaces) with input = {}'.format(
                [workspace.name for workspace in workspaces]
            ))
            return workspaces

        result = list(filter(
            lambda workspace: identifier == workspace.identifier,
            workspaces
        ))

        if not result:
            logger.warning('no workspaces found with identifier = {} and with input = {}'.format(
                identifier, [workspace.name for workspace in workspaces]
            ))
        return result



import click
from tabulate import tabulate
from typing import List, Optional

from togglcmder.toggl.cli.helpers import retrieve_cache_from_context
from togglcmder.toggl.cli.helpers import retrieve_downloader_from_context

from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.views.workspace import Workspace as WorkspaceView


def sync_or_retrieve_workspaces(context_obj: dict) -> List[Workspace]:
    caching = retrieve_cache_from_context(context_obj)
    downloader = retrieve_downloader_from_context(context_obj)

    # Download any new workspaces added remotely (this is only possible via
    # the web interface anyway), if remote sync is enabled.
    if context_obj['sync'] is True:
        current_workspaces = downloader.download_workspaces()
        if current_workspaces:
            caching.update_workspace_cache(current_workspaces)
    else:
        # Otherwise we just pull from the local cache.
        current_workspaces = caching.retrieve_workspace_cache()

    return current_workspaces


def retrieve_workspace_from_context(context: click.Context) -> Optional[Workspace]:
    workspace = context.obj['data']['workspace']
    return workspace


@click.command(
    help='List the currently available workspaces.'
)
@click.pass_context
def workspaces(context: click.Context):
    current_workspaces = sync_or_retrieve_workspaces(context.obj)

    # This means the user is new and has not synced anything from the Toggl
    # servers. There is always at least one workspace called Everything!
    if not current_workspaces:
        click.echo(click.style('WARNING', fg='red') +
                   ': There are no workspaces available.')
        return

    # The only thing we can do with workspaces is list them,
    # so create the view and use the echo functionality click
    # provides so we aren't listing workspaces in the log.
    workspace_view = WorkspaceView(current_workspaces)
    click.echo_via_pager(tabulate(
        workspace_view.values(),
        headers=WorkspaceView.headers(),
        tablefmt="grid"))



from __future__ import annotations
from datetime import datetime
from tzlocal import get_localzone
from typing import Optional

from togglcmder.toggl.types.workspace import Workspace


class WorkspaceBuilder(object):
    def __init__(self, workspace: Optional[Workspace] = None):
        if workspace is not None:
            self.__identifier = workspace.identifier
            self.__name = workspace.name
            self.__last_updated = workspace.last_updated
        else:
            self.__identifier = None
            self.__name = None
            self.__last_updated = None

    def identifier(self, identifier: int) -> WorkspaceBuilder:
        self.__identifier = identifier
        return self

    def name(self, name: str) -> WorkspaceBuilder:
        self.__name = name
        return self

    def last_updated(self, *, last_update: Optional[str] = None,
                     epoch: Optional[int] = None) -> WorkspaceBuilder:
        if last_update:
            self.__last_updated = self.__datetime_from_str(last_update)
        elif epoch:
            self.__last_updated = self.__datetime_from_timestamp(epoch)
        return self

    def build(self) -> Workspace:
        return Workspace(
            identifier=self.__identifier,
            name=self.__name,
            last_updated=self.__last_updated)

    @staticmethod
    def __datetime_from_str(date: str) -> datetime:
        parsed_date = datetime.fromisoformat(date.replace('Z', '+00:00'))
        if not parsed_date.tzinfo:
            return get_localzone().localize(parsed_date)
        return parsed_date.astimezone(get_localzone())

    @staticmethod
    def __datetime_from_timestamp(timestamp: int) -> datetime:
        return get_localzone().localize(datetime.fromtimestamp(timestamp))



import requests
from requests.auth import HTTPBasicAuth
import json
import logging
from typing import List

from togglcmder.toggl.endpoints.api import API

from togglcmder.toggl.types.project import Project
from togglcmder.toggl.encoders.project_encoder import ProjectEncoder
from togglcmder.toggl.decoders.project_decoder import ProjectDecoder

from togglcmder.toggl.types.tag import Tag
from togglcmder.toggl.encoders.tag_encoder import TagEncoder
from togglcmder.toggl.decoders.tag_decoder import TagDecoder

from togglcmder.toggl.types.time_entry import TimeEntry
from togglcmder.toggl.encoders.time_entry_encoder import TimeEntryEncoder
from togglcmder.toggl.decoders.time_entry_decoder import TimeEntryDecoder


class Commands(object):
    def __init__(self, auth: str):
        self.__auth = HTTPBasicAuth(auth, 'api_token')

    @staticmethod
    def __log_command(logger: logging.Logger, result: requests.Response):
        logger.debug(result.request.body)
        logger.debug(result.text)

    def add_tag(self, tag: Tag) -> Tag:
        result = requests.post(
            API().tags,
            data=json.dumps(tag,
                            cls=TagEncoder),
            auth=self.__auth)

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=TagDecoder)

    def delete_tag(self, tag: Tag) -> None:
        result = requests.delete(
            API().tags.details(tag.identifier),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()

    def update_tag(self, tag: Tag) -> Tag:
        result = requests.put(
            API().tags.details(tag.identifier),
            data=json.dumps(tag,
                            cls=TagEncoder),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=TagDecoder)

    def add_project(self, project: Project) -> Project:
        result = requests.post(
            API().projects,
            data=json.dumps(project,
                            cls=ProjectEncoder),
            auth=self.__auth)

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=ProjectDecoder)

    def delete_project(self, project: Project) -> None:
        result = requests.delete(
            API().projects.details(project.identifier),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()

    def delete_projects(self, projects: List[Project]) -> None:
        result = requests.delete(
            API().projects.details(
                identifiers=[project.identifier for project in projects]
            ),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()

    def update_project(self, project: Project) -> Project:
        result = requests.put(
            API().projects.details(project.identifier),
            data=json.dumps(project,
                            cls=ProjectEncoder),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=ProjectDecoder)

    def start_time_entry(self, time_entry: TimeEntry) -> TimeEntry:
        result = requests.post(
            API().time_entries.start(),
            data=json.dumps(time_entry,
                            cls=TimeEntryEncoder),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=TimeEntryDecoder)

    def stop_time_entry(self, time_entry: TimeEntry) -> TimeEntry:
        result = requests.put(
            API().time_entries.stop(time_entry.identifier),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=TimeEntryDecoder)

    def add_completed_time_entry(self, time_entry: TimeEntry) -> TimeEntry:
        result = requests.post(
            API().time_entries,
            data=json.dumps(time_entry,
                            cls=TimeEntryEncoder),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=TimeEntryDecoder)

    def update_completed_time_entry(self, time_entry: TimeEntry) -> TimeEntry:
        result = requests.put(
            API().time_entries.details(time_entry.identifier),
            data=json.dumps(time_entry,
                            cls=TimeEntryEncoder),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=TimeEntryDecoder)

    def delete_time_entry(self, time_entry: TimeEntry) -> None:
        result = requests.delete(
            API().time_entries.details(time_entry.identifier),
            auth=self.__auth
        )

        Commands.__log_command(logging.getLogger(__name__), result)

        result.raise_for_status()

    def reset_api_token(self) -> str:
        reply = requests.post(API().users.reset_api_token(),
                              auth=self.__auth
                              )
        reply.raise_for_status()
        return reply.text



from json import JSONDecoder

from togglcmder.toggl.builders.time_entry_builder import TimeEntryBuilder


class TimeEntryDecoder(JSONDecoder):
    def __init__(self, *args: tuple, **kwargs: dict):
        JSONDecoder.__init__(self,
                             object_hook=TimeEntryDecoder.object_hook,
                             *args,
                             **kwargs)

    @staticmethod
    def object_hook(obj: dict):
        if 'data' in obj:
            return obj['data']

        if 'id' in obj:
            return TimeEntryBuilder()\
                .identifier(obj['id'])\
                .description(obj.get('description', None))\
                .workspace_identifier(obj['wid'])\
                .project_identifier(obj.get('pid', None))\
                .start_time(start_time=obj['start'])\
                .duration(obj.get('duration', None))\
                .stop_time(stop_time=obj.get('stop', None))\
                .tags(obj.get('tags', None))\
                .last_updated(last_update=obj.get('at', None))\
                .build()

        return obj



from typing import List

import requests, json
from requests.auth import HTTPBasicAuth
from datetime import datetime
import logging

from togglcmder.toggl.endpoints.api import API

from togglcmder.toggl.types.user import User
from togglcmder.toggl.decoders.user_decoder import UserDecoder

from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.decoders.workspace_decoder import WorkspaceDecoder

from togglcmder.toggl.types.tag import Tag
from togglcmder.toggl.decoders.tag_decoder import TagDecoder

from togglcmder.toggl.types.project import Project
from togglcmder.toggl.decoders.project_decoder import ProjectDecoder

from togglcmder.toggl.types.time_entry import TimeEntry
from togglcmder.toggl.decoders.time_entry_decoder import TimeEntryDecoder

class Downloader(object):
    def __init__(self, auth: str):
        self.__auth = HTTPBasicAuth(auth, 'api_token')

    @staticmethod
    def __log_download(logger: logging.Logger, result: requests.Response):
        logger.debug(result.request.body)
        logger.debug(result.text)

    def download_user_data(self) -> User:
        reply = requests.get(
            API().users.details(),
            auth=self.__auth
        )

        # Commented this out because it may show sensitive information.
        # Downloader.__log_download(logging.getLogger(__name__), reply)

        reply.raise_for_status()
        return json.loads(reply.text,
                          cls=UserDecoder)

    def download_workspaces(self) -> List[Workspace]:
        reply = requests.get(
            API().workspaces,
            auth=self.__auth
        )

        Downloader.__log_download(logging.getLogger(__name__), reply)

        reply.raise_for_status()
        return json.loads(reply.text,
                          cls=WorkspaceDecoder)

    def download_tags(self, workspace: Workspace) -> List[Tag]:
        reply = requests.get(
            API().workspaces.tags(workspace.identifier),
            auth=self.__auth
        )

        Downloader.__log_download(logging.getLogger(__name__), reply)

        reply.raise_for_status()
        return json.loads(reply.text,
                          cls=TagDecoder)

    def download_projects(self, workspace: Workspace) -> List[Project]:
        reply = requests.get(
            API().workspaces.projects(workspace.identifier),
            auth=self.__auth
        )

        Downloader.__log_download(logging.getLogger(__name__), reply)

        reply.raise_for_status()
        return json.loads(reply.text,
                          cls=ProjectDecoder)

    def download_time_entries(self, start: datetime, end: datetime) -> List[TimeEntry]:
        reply = requests.get(
            API().time_entries.search(start, end),
            auth=self.__auth
        )

        Downloader.__log_download(logging.getLogger(__name__), reply)

        reply.raise_for_status()
        return json.loads(reply.text,
                          cls=TimeEntryDecoder)

    def get_current_time_entry(self) -> TimeEntry:
        result = requests.get(
            API().time_entries.current(),
            auth=self.__auth
        )

        Downloader.__log_download(logging.getLogger(__name__), result)

        result.raise_for_status()
        return json.loads(result.text,
                          cls=TimeEntryDecoder)



from typing import Optional
from datetime import datetime
from functools import total_ordering


@total_ordering
class User(object):
    def __init__(self, *, name: str, api_token: str,
                 identifier: Optional[int] = None,
                 last_updated: Optional[datetime] = None):
        self.__name = name
        self.__api_token = api_token
        self.__identifier = identifier
        self.__last_updated = last_updated

    def __eq__(self, other):
        if not isinstance(other, User):
            return NotImplemented

        return self.__name == other.name \
            and self.__api_token == other.api_token \
            and self.__identifier == other.identifier \
            and self.__last_updated == other.last_updated

    def __lt__(self, other):
        if not isinstance(other, User):
            return NotImplemented
        return self.__name.lower() < other.__name.lower()

    def __str__(self):
        return "{},{},{},{}".format(
            self.name,
            self.identifier,
            self.api_token,
            self.last_updated.isoformat())

    @property
    def name(self) -> str:
        return self.__name

    @property
    def api_token(self) -> str:
        return self.__api_token

    @property
    def identifier(self) -> int:
        return self.__identifier

    @property
    def last_updated(self) -> datetime:
        return self.__last_updated



from datetime import datetime

from typing import List, Optional
from functools import total_ordering


@total_ordering
class TimeEntry(object):
    def __init__(self, *, description: str, start_time: datetime,
                 identifier: Optional[int] = None,
                 project_identifier: Optional[int] = None,
                 workspace_identifier: Optional[int] = None,
                 stop_time: Optional[datetime] = None,
                 duration: Optional[int] = None,
                 tags: Optional[List[str]] = None,
                 last_updated: Optional[datetime] = None):
        self.__description = description
        self.__start_time = start_time
        self.__identifier = identifier
        self.__project_identifier = project_identifier
        self.__workspace_identifier = workspace_identifier
        self.__stop_time = stop_time
        self.__duration = duration
        self.__tags = tags
        self.__last_updated = last_updated

    def __eq__(self, other):
        if not isinstance(other, TimeEntry):
            return NotImplemented

        return self.__description == other.description \
            and self.__start_time == other.start_time \
            and self.__identifier == other.identifier \
            and self.__project_identifier == other.project_identifier \
            and self.__workspace_identifier == other.workspace_identifier \
            and self.__stop_time == other.stop_time \
            and self.__duration == other.duration \
            and self.__tags == other.tags \
            and self.__last_updated == other.last_updated

    def __lt__(self, other):
        if not isinstance(other, TimeEntry):
            raise NotImplemented
        return self.__start_time < other.start_time

    def __str__(self):
        return "{},{},{},{},{},{},{},{},{}".format(
            self.description,
            self.start_time.isoformat(),
            self.stop_time.isoformat() if self.stop_time else None,
            self.duration,
            self.identifier,
            self.project_identifier,
            self.workspace_identifier,
            self.tags,
            self.last_updated)

    @property
    def description(self) -> str:
        return self.__description

    @property
    def start_time(self) -> datetime:
        return self.__start_time

    @property
    def identifier(self) -> int:
        return self.__identifier

    @property
    def project_identifier(self) -> int:
        return self.__project_identifier

    @property
    def workspace_identifier(self) -> int:
        return self.__workspace_identifier

    @property
    def stop_time(self) -> datetime:
        return self.__stop_time

    @property
    def duration(self) -> int:
        return self.__duration

    @property
    def tags(self) -> List[str]:
        return self.__tags

    @property
    def last_updated(self) -> datetime:
        return self.__last_updated



import click
from tabulate import tabulate
import logging
from typing import List

from togglcmder.toggl.cli.workspaces \
    import sync_or_retrieve_workspaces, retrieve_workspace_from_context

from togglcmder.toggl.cli.helpers import retrieve_cache_from_context
from togglcmder.toggl.cli.helpers import retrieve_commands_from_context
from togglcmder.toggl.cli.helpers import retrieve_downloader_from_context

from togglcmder.toggl.caching import Caching
from togglcmder.toggl.commands import Commands

from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.filters.workspaces import Workspaces as WorkspaceFilter

from togglcmder.toggl.types.tag import Tag
from togglcmder.toggl.builders.tag_builder import TagBuilder
from togglcmder.toggl.views.tag import Tag as TagView
from togglcmder.toggl.filters.tags import Tags as TagFilter


def sync_or_retrieve_tags(context_obj: dict, workspace: Workspace) -> List[Tag]:
    caching = retrieve_cache_from_context(context_obj)
    downloader = retrieve_downloader_from_context(context_obj)

    # Download any new tags if remote sync is enabled.
    if context_obj['sync'] is True:
        current_tags = downloader.download_tags(workspace)
        if current_tags:
            caching.update_tag_cache(current_tags)
    else:
        # Otherwise we just pull from the local cache.
        current_tags = caching.retrieve_tag_cache()

        if current_tags:
            # Retrieval from the DB doesn't filter on workspace, so we do that ourselves.
            current_tags = TagFilter.filter_on_workspace(current_tags, workspace)

    return current_tags


@click.group(
    help='Add, update, delete, and list tags.'
)
@click.option('--workspace')
@click.pass_context
def tags(context: click.Context, workspace: str):
    workspace_name = workspace or context.obj['config']['default_workspace']

    workspaces = sync_or_retrieve_workspaces(context.obj)

    if workspace_name:
        workspaces = WorkspaceFilter.filter_on_name(
            workspaces,
            workspace_name
        )

        workspaces_found = len(workspaces)
        if workspaces_found == 1:
            context.obj['data']['workspace'] = workspaces[0]

    context.obj['data']['workspaces'] = workspaces

    current_tags = []
    for current_workspace in workspaces:
        synced_tags = sync_or_retrieve_tags(context.obj, current_workspace)
        if synced_tags:
            current_tags.extend(synced_tags)

    context.obj['data']['tags'] = current_tags


@tags.command('add')
@click.option('--name',
              required=True)
@click.pass_context
def tag_add(context: click.Context, name: str):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace required to add a new tag.
        click.echo(click.style('ERROR', fg='red') +
                   ': a workspace is required when adding a new tag.')
        return

    current_tags = context.obj['data']['tags']
    if current_tags:
        current_tags = TagFilter.filter_on_name(
            current_tags, name)
        if current_tags:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': tag({name}) with this name already'
                       f' exists in this workspace({workspace.name}).')
            return

    tag = TagBuilder()\
        .name(name)\
        .workspace_identifier(workspace.identifier)\
        .build()

    cache = retrieve_cache_from_context(context.obj)
    commands = retrieve_commands_from_context(context.obj)

    # This raises if there is any issues, which means it doesn't get added
    # to the local cache.
    try:
        added_tag = commands.add_tag(tag)

        # The new tag was added to the remote Toggl servers, so now
        # we can add it to our local cache as well.
        cache.update_tag_cache([added_tag])
        click.echo(click.style('SUCCESS', fg='green') +
                   f': added new tag({added_tag.name}) to'
                   f' workspace({workspace.name})!')

    except Exception as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to add new tag({name}) to workspace({workspace.name}).'
                   ' An exception has been logged; check the logs for more information.')


@tags.command('delete')
@click.option('--name')
@click.option('--multiple',
              is_flag=True,
              default=False,
              show_default=True)
@click.pass_context
def tag_delete(context: click.Context, name: str, multiple: bool):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when deleting a tag.')
        return

    current_tags = context.obj['data']['tags']

    if not current_tags:
        click.echo(click.style('WARNING') +
                   f': there are no tags in this workspace({workspace.name}).')
        return

    if name:
        current_tags = TagFilter.filter_on_name(
            current_tags, name
        )
        if not current_tags:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no tags exist with this name({name}'
                       f' in this workspace({workspace.name})')
            return

    if not multiple and len(current_tags) > 1:
        click.echo(click.style('ERROR', fg='red') +
                   ': multiple tags found matching the criteria, but --multiple '
                   'not specified.. please use the --multiple option or tighten '
                   'the search criteria.')
        return

    cache = retrieve_cache_from_context(context.obj)
    commands = retrieve_commands_from_context(context.obj)

    # This raises if there is any issues, which means it doesn't get added
    # to the local cache.
    try:
        for tag in current_tags:
            commands.delete_tag(tag)

            # The new tag was removed from the remote Toggl servers, so now
            # we can remove it from our local cache as well.
            cache.remove_tag_from_cache(tag)
            click.echo(click.style('SUCCESS', fg='green') +
                       f': deleted tag({tag.name})!')

    except Exception as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to delete the tag(s). An exception has been logged;'
                   ' check the logs for more information.')


@tags.command('update')
@click.option('--old-name',
              required=True)
@click.option('--new-name',
              required=True)
@click.pass_context
def tag_update(context: click.Context, old_name: str, new_name: str):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when updating a tag.')
        return

    current_tags = context.obj['data']['tags']
    if not current_tags:
        click.echo(click.style('WARNING', fg='yellow') +
                   f': no tags to update in this workspace({workspace.name}).')
        return

    current_tags = TagFilter.filter_on_name(current_tags, old_name)
    if not current_tags:
        click.echo(click.style('WARNING', fg='yellow') +
                   f': no tags with name({old_name}) found in this workspace({workspace.name}).')
        return
    elif len(current_tags) > 1:
        click.echo(click.style('ERROR', fg='red') +
                   f': multiple tags found matching name({old_name}), in'
                   f' this workspace({workspace.name}).. please be more specific.')
        return

    # Only support updating a single tag at a time, because updating multiple tags
    # at once doesn't make sense when the only thing you can update is the name.
    builder = TagBuilder(current_tags[0])
    builder.name(new_name)
    updated_tag = builder.build()

    cache = context.obj['cache']
    assert(isinstance(cache, Caching))

    commands = context.obj['commands']
    assert(isinstance(commands, Commands))

    try:
        updated_tag = commands.update_tag(updated_tag)
        cache.update_tag_cache([updated_tag])
        click.echo(click.style('SUCCESS', fg='green') +
                   f': updated tag({old_name}) to tag({new_name})!')

    except Exception as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to update the tags. An exception has been logged;'
                   ' check the logs for more information.')


@tags.command('list')
@click.option('--name')
@click.option('--sort-by',
              type=click.Choice([val.__str__() for val in TagView.headers()]),
              default=TagView.headers()[0],
              show_default=True)
@click.pass_context
def tag_list(context: click.Context, name: str, sort_by: str):
    workspaces = context.obj['data']['workspaces']
    current_tags = context.obj['data']['tags']
    if not current_tags:
        click.echo(click.style('WARNING', fg='yellow') +
                   f': no tags found in the specified workspace(s).')
        return

    if name:
        current_tags = TagFilter.filter_on_name(
            current_tags, name
        )
        if not current_tags:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no tags found with name({name}).')
            return

    rows = []
    for current_workspace in workspaces:
        filtered_tags = TagFilter.filter_on_workspace(
            current_tags, current_workspace
        )
        if filtered_tags:
            rows.extend(
                TagView(
                    tags=filtered_tags,
                    workspace=current_workspace
                ).values()
            )

    sort_by_index = 0
    if sort_by:
        if sort_by not in TagView.headers():
            click.echo(f'Choose a valid header to sort by: {TagView.headers()}')
            return
        for index, header in enumerate(TagView.headers()):
            if sort_by == header:
                sort_by_index = index
                break

    rows = sorted(rows, key=lambda key: key[sort_by_index], reverse=True)
    click.echo_via_pager(tabulate(
        rows,
        headers=TagView.headers(),
        tablefmt="grid"
    ))



#!/usr/bin/env python

import os
import click
import json
import logging
import logging.handlers
import sys

from typing import Optional, List

from togglcmder.version import __version__

from togglcmder.toggl.cli.workspaces import workspaces as workspace_cli
from togglcmder.toggl.cli.projects import projects as project_cli
from togglcmder.toggl.cli.tags import tags as tag_cli
from togglcmder.toggl.cli.time_entries import timers as timer_cli

from togglcmder.toggl.caching import Caching
from togglcmder.toggl.downloader import Downloader
from togglcmder.toggl.commands import Commands


@click.group(invoke_without_command=True)
@click.option(
    '--api-key',
    help='Your API key for the Toggl API',
    type=str,
    show_envvar=True,
    allow_from_autoenv=True
)
@click.option(
    '--reset-api-key',
    help='Resets the Toggl API key and downloads it to the local device.',
    is_flag=True,
    flag_value=True,
    callback=lambda c, x, v: click.confirm('Are you sure you want to reset the API key?') if v else False
)
@click.option(
    '--default-workspace',
    help='Specify a default workspace to use when otherwise not specified.',
    show_envvar=True,
    allow_from_autoenv=True
)
@click.option(
    '--default-project',
    help='Specify a default project to use when otherwise not specified.',
    show_envvar=True,
    allow_from_autoenv=True
)
@click.option(
    '--default-tags',
    help='Specify some default tags to always include for timers started on this machine.',
    show_envvar=True,
    allow_from_autoenv=True
)
@click.option(
    '--default-time-entry-start-days',
    help='Specify some default start time to refresh time entries.',
    type=int
    # help='This is the number of days before now we should start looking for time '
    #     'entries when doing a refresh. The default of "30" means look for '
    #     'time entries starting from 30 days ago.'
)
@click.option(
    '--default-time-entry-stop-days',
    help='Specify some default stop time to refresh time entries.',
    type=int
    # help='This is the number of days before now that we should stop looking '
    #     'for time entries when doing a refresh. The default of "0" means'
    #     'make the stop time now.'
)
@click.option(
    '--verbosity', '-v',
    default=0,
    count=True
)
@click.option(
    '--version',
    is_flag=True,
    callback=lambda c, t, v: click.echo(f"Version {__version__}") & c.exit() if v else False,
    expose_value=False,
    is_eager=True
)
@click.option(
    '--sync',
    help='Download from remote Toggl servers before attempting item lookups.',
    is_flag=True,
    default=False,
    show_default=True
)
@click.option(
    '--show-config',
    help='Simply prints the current configuration.',
    is_flag=True,
    default=False
)
@click.pass_context
def main(context: click.Context,
         api_key: str,
         reset_api_key: bool,
         default_workspace: str,
         default_project: str,
         default_tags: str,
         default_time_entry_start_days: int,
         default_time_entry_stop_days: int,
         verbosity: int,
         sync: bool,
         show_config: bool):

    # This object (context object provided by Click for sharing data between the
    # various command chains) contains configuration, defaults, and the objects we
    # create in the entry point to establish a connection to the database and start
    # logging to the file. The data key contains the current relevant objects, such
    # as workspaces, projects, tags, or time entries.
    context.obj = {
        'config': {
            'api_key': Optional[str],
            'default_workspace': Optional[str],
            'default_project': Optional[str],
            'default_tags': Optional[List[str]],
            'default_time_entry_window_start_days': 5,
            'default_time_entry_window_stop_days': 0
        },
        'cache': Optional[Caching],
        'downloader': Optional[Downloader],
        'commands': Optional[Commands],
        'sync': sync,
        'data': {
            'workspaces': [],
            'workspace': None,
            'projects': [],
            'project': None,
            'tags': [],
            'time_entries': []
        }
    }
    context.obj['config']['api_key'] = None
    context.obj['config']['default_workspace'] = None
    context.obj['config']['default_project'] = None
    context.obj['config']['default_tags'] = None

    # Set up the paths paths that we need.
    app_dir = click.get_app_dir('togglcmder')

    # Configuration file is automatically created and stored in the application
    # directory for the user. If it exists, it's simply loaded.
    config = os.path.join(app_dir, 'toggl.json')
    if not os.path.exists(app_dir):
        os.mkdir(app_dir)

    if not os.path.exists(config):
        with open(config, 'w') as handle:
            json.dump(context.obj['config'], handle, sort_keys=True, indent=4)
    else:
        with open(config, 'r') as handle:
            # Merge the defaults in with the actual configuration.
            context.obj['config'] = {**context.obj['config'], **json.load(handle)}

    # If the user provides the API key, we overwrite the key from the configuration
    # file and use that for all future requests.
    if not api_key and not context.obj['config']['api_key']:
        click.echo(click.style('ERROR', fg='red', bold=True) +
                   ': there is no API key yet.. please insert the key into'
                   f' {config} or provide it with the "--api-key" argument!')
        return

    logger = logging.getLogger()

    # Logs to the app directory instead of where the script is run from.
    log_path = os.path.join(app_dir, 'toggl.log')
    log_file_handle = logging.handlers.RotatingFileHandler(log_path)

    # TODO: is this formatted string sufficient for logging? Too much?
    formatter = logging.Formatter(
        "%(asctime)s: %(levelname)s: %(name)s: %(lineno)d: %(message)s",
        "%Y-%m-%dT%H:%M:%S")

    log_file_handle.setFormatter(formatter)

    logger.addHandler(log_file_handle)

    # The way logger handles the debug levels is a bit odd, so we need to do
    # this to allow the the user to change the levels as expected. e.g. -vvv
    # allows more verbosity than just -v.
    logger.setLevel(60 - ((3 + verbosity) * 10))

    cache = Caching(cache_name=os.path.join(app_dir, 'cache.db'))
    context.obj['cache'] = cache

    if api_key:
        context.obj['config']['api_key'] = api_key

    # We obviously cannot do anything without an API key.
    if not context.obj['config']['api_key']:
        click.echo(click.style("ERROR", fg="red") +
                   ": there is no API key defined. Check the README or Wiki for "
                   "more information.")
        return

    # The API key can be reset and is automatically updated in both the file and
    # the running instance of the script.
    if reset_api_key:
        try:
            context.obj['config']['api_key'] = Commands(
                context.obj['config']['api_key']).reset_api_token().strip('"')
            click.echo(click.style("SUCCESS", fg="green") +
                       "API key successfully reset!")
        except Exception as e:
            click.echo(click.style("ERROR", fg="red") +
                       "Failed to reset API key. Check logs for more information.")
            logger.error(e)

    # These defaults can allow users to start timers without needing to specify
    # a workspace and project every single time. If the specific options for a
    # given command are not provided then these are used.
    if default_workspace:
        if default_workspace == 'None':
            context.obj['config']['default_workspace'] = None
        else:
            context.obj['config']['default_workspace'] = default_workspace

    if default_project:
        if default_project == 'None':
            context.obj['config']['default_project'] = None
        else:
            context.obj['config']['default_project'] = default_project

    if default_tags:
        if default_tags == 'None':
            context.obj['config']['default_tags'] = None
        else:
            context.obj['config']['default_tags'] = [t.strip() for t in default_tags.split(',')]

    if default_time_entry_start_days:
        context.obj['config']['default_time_entry_window_start_days'] = default_time_entry_start_days

    if default_time_entry_stop_days:
        context.obj['config']['default_time_entry_window_stop_days'] = default_time_entry_stop_days

    # We've updated the configuration internally, now update it on disk.
    with open(config, 'w') as handle:
        json.dump(context.obj['config'], handle, sort_keys=True, indent=4)

    if show_config:
        click.echo(context.obj['config'])

    # Set up the command and downloader objects with the API key.
    context.obj['commands'] = Commands(context.obj['config']['api_key'])
    context.obj['downloader'] = Downloader(context.obj['config']['api_key'])


######################################################


assert(isinstance(main, click.Group))

main.add_command(workspace_cli)
main.add_command(project_cli)
main.add_command(tag_cli)
main.add_command(timer_cli)

######################################################

if __name__ == "__main__":
    main(auto_envvar_prefix='TOGGL')



from datetime import datetime
import re as regex
from typing import List
import logging

from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.types.project import Project
from togglcmder.toggl.types.tag import Tag
from togglcmder.toggl.types.time_entry import TimeEntry


class TimeEntries(object):

    @staticmethod
    def filter_single(
            time_entries: List[TimeEntry],
    ) -> TimeEntry:
        if not time_entries:
            raise RuntimeError('no time entries passed in when single time entry expected')
        elif len(time_entries) > 1:
            raise RuntimeError('multiple time entries passed in when single time entry expected')

        return time_entries[0]

    @staticmethod
    def filter_on_description(
            time_entries: List[TimeEntry],
            description: str
    ) -> List[TimeEntry]:
        logger = logging.getLogger(__name__)
        if not description:
            logger.debug('no description provided for filter_on_description(time_entries) with input = {}'.format(
                [entry.description for entry in time_entries]
            ))
            return time_entries

        result = list(filter(
            lambda time_entry:
                regex.match("^{}$".format(description),
                            time_entry.description if time_entry.description else "",
                            regex.IGNORECASE),
            time_entries
        ))

        if not result:
            logger.warning('no time entries found with description = {} and with input = {}'.format(
                description, [entry.description for entry in time_entries]
            ))
        return result

    @staticmethod
    def filter_on_workspace(
            time_entries: List[TimeEntry],
            workspace: Workspace
    ) -> List[TimeEntry]:
        logger = logging.getLogger(__name__)
        if not workspace:
            logger.debug('no workspace provided for filter_on_workspace(time_entries) with input = {}'.format(
                [(entry.description, entry.workspace_identifier) for entry in time_entries]
            ))
            return time_entries

        result = list(filter(
            lambda time_entry: workspace.identifier == time_entry.workspace_identifier,
            time_entries
        ))

        if not result:
            logger.warning('no time entries found with workspace = {} and with input = {}'.format(
                (workspace.name, workspace.identifier),
                [(entry.description, entry.workspace_identifier) for entry in time_entries]
            ))
        return result

    @staticmethod
    def filter_on_project(
            time_entries: List[TimeEntry],
            project: Project
    ) -> List[TimeEntry]:
        logger = logging.getLogger(__name__)
        if not project:
            logger.debug('no project provided for filter_on_project(time_entries) with input = {}'.format(
                [(entry.description, entry.project_identifier) for entry in time_entries]
            ))
            return time_entries

        result = list(filter(
            lambda time_entry: project.identifier == time_entry.project_identifier,
            time_entries
        ))

        if not result:
            logger.warning('no time entries found with project = {} and with input = {}'.format(
                (project.name, project.identifier),
                [(entry.description, entry.project_identifier) for entry in time_entries]
            ))
        return result

    @staticmethod
    def filter_on_all_tags(
            time_entries: List[TimeEntry],
            tags: List[Tag]
    ) -> List[TimeEntry]:
        logger = logging.getLogger(__name__)
        if not tags:
            logger.debug('no tags provided for filter_on_all_tags(time_entries) with input = {}'.format(
                [entry.tags for entry in time_entries]
            ))
            return time_entries

        result = list(filter(
            lambda time_entry: set(
                [tag.name for tag in tags]
            ).issubset(time_entry.tags),
            time_entries
        ))

        if not result:
            logger.warning('no time entries found with tags = {} and with input = {}'.format(
                [tag.name for tag in tags], [entry.tags for entry in time_entries]
            ))
        return result

    @staticmethod
    def filter_on_any_tags(
            time_entries: List[TimeEntry],
            tags: List[Tag]
    ) -> List[TimeEntry]:
        logger = logging.getLogger(__name__)
        if not tags:
            logger.debug('no tags provided for filter_on_any_tags(time_entries) with input = {}'.format(
                [entry.tags for entry in time_entries]
            ))
            return time_entries

        result = list(filter(
            lambda time_entry:
                not set([tag.name for tag in tags]).isdisjoint(time_entry.tags),
            time_entries
        ))

        if not result:
            logger.warning('no time entries found with tags = {} and with input = {}'.format(
                [tag.name for tag in tags], [entry.tags for entry in time_entries]
            ))
        return result

    @staticmethod
    def filter_on_date_range(
            time_entries: List[TimeEntry],
            start: datetime,
            end: datetime
    ) -> List[TimeEntry]:
        logger = logging.getLogger(__name__)
        if not start and not end:
            logger.debug('neither start nor end provided for filter_on_date_range(time_entries) with input = {}'.format(
                [(entry.start_time.isoformat(), entry.stop_time.isoformat()) for entry in time_entries]
            ))
            return time_entries

        result = None
        if start and end:
            result = list(filter(
                lambda time_entry: start < time_entry.start_time < end,
                time_entries
            ))
            if not result:
                logger.warning('no time entries found for start = {} and end = {} and with input = {}'.format(
                    start.isoformat(), end.isoformat(),
                    [entry.start_time.isoformat() for entry in time_entries]
                ))

        elif start:
            result = list(filter(
                lambda time_entry: start < time_entry.start_time,
                time_entries
            ))
            if not result:
                logger.warning('no time entries found with start = {} and with input = {}'.format(
                    start.isoformat(), [entry.start_time.isoformat() for entry in time_entries]
                ))

        elif end:
            result = list(filter(
                lambda time_entry: time_entry.start_time < end,
                time_entries
            ))
            if not result:
                logger.warning('no time entries found with end = {} and with input = {}'.format(
                    end.isoformat(), [entry.stop_time.isoformat() for entry in time_entries]
                ))

        return result



"""

"""


class Workspaces(object):
    """

    """

    def __init__(self, url: str):
        """

        :param url:
        """
        self.__url = url + "/workspaces"

    def __repr__(self):
        return self.__url

    def details(self, id_: int) -> str:
        """

        :param id_:
        :return:
        """
        return self.__url + "/{}".format(id_)

    def projects(self, id_: int) -> str:
        return self.__url + "/{}/projects".format(id_)

    def tags(self, id_: int) -> str:
        return self.__url + "/{}/tags".format(id_)



import unittest
from datetime import datetime

from togglcmder.toggl.filters.time_entries import TimeEntries
from togglcmder.toggl.filters.tags import Tags
from togglcmder.toggl.filters.workspaces import Workspaces
from togglcmder.toggl.filters.projects import Projects

from togglcmder.toggl.types.project import Project
from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.types.tag import Tag
from togglcmder.toggl.types.time_entry import TimeEntry
from togglcmder.toggl.types.user import User


class TestFilters(unittest.TestCase):
    WORKSPACE_ONE = Workspace(
        name='Test Workspace One',
        identifier=1,
        last_updated=datetime.now())

    WORKSPACE_TWO = Workspace(
        name='Test Workspace Two',
        identifier=2,
        last_updated=datetime.now())

    PROJECT_ONE = Project(
        name='Test Project One',
        color=Project.Color.RED,
        last_updated=datetime.now(),
        created=datetime.now(),
        identifier=1,
        workspace_identifier=1)

    PROJECT_TWO = Project(
        name='Test Project Two',
        color=Project.Color.BLACK,
        last_updated=datetime.now(),
        created=datetime.now(),
        identifier=2,
        workspace_identifier=2)

    TAG_ONE = Tag(
        name='Test Tag One',
        identifier=1,
        workspace_identifier=1)

    TAG_TWO = Tag(
        name='Test Tag Two',
        identifier=2,
        workspace_identifier=1
    )

    TAG_THREE = Tag(
        name='Test Tag Three',
        identifier=3,
        workspace_identifier=1
    )

    TIME_ENTRY_ONE = TimeEntry(
        description='Test Entry One',
        start_time=datetime.now(),
        identifier=1,
        duration=30,
        project_identifier=1,
        workspace_identifier=1,
        stop_time=datetime.now(),
        tags=['Test Tag One', 'Test Tag Two'],
        last_updated=datetime.now()
    )

    TIME_ENTRY_TWO = TimeEntry(
        description='Test Entry Two',
        start_time=datetime.now(),
        identifier=2,
        duration=60,
        project_identifier=1,
        workspace_identifier=1,
        stop_time=datetime.now(),
        tags=['Test Tag Three'],
        last_updated=datetime.now()
    )

    USER = User(
        name='Test User',
        api_token='1234',
        identifier=1,
        last_updated=datetime.now())

    def test_filter_on_workspace_id(self) -> None:
        self.assertEqual([TestFilters.WORKSPACE_ONE],
                         Workspaces.filter_on_identifier(
                             [TestFilters.WORKSPACE_ONE, TestFilters.WORKSPACE_TWO],
                             TestFilters.WORKSPACE_ONE.identifier
                         ))

    def test_filter_projects_on_workspace(self) -> None:
        self.assertEqual([TestFilters.PROJECT_ONE],
                         Projects.filter_on_workspace(
                             [TestFilters.PROJECT_ONE, TestFilters.PROJECT_TWO],
                             TestFilters.WORKSPACE_ONE
                         ))

    def test_filter_projects_on_name(self) -> None:
        self.assertEqual([TestFilters.PROJECT_ONE],
                         Projects.filter_on_name(
                             [TestFilters.PROJECT_ONE, TestFilters.PROJECT_TWO],
                             TestFilters.PROJECT_ONE.name
                         ))

    def test_filter_tags_on_workspace(self) -> None:
        self.assertEqual([],
                         Tags.filter_on_workspace(
                             [TestFilters.TAG_ONE, TestFilters.TAG_TWO, TestFilters.TAG_THREE],
                             TestFilters.WORKSPACE_TWO
                         ))

    def test_filter_tags_on_name(self) -> None:
        self.assertEqual([TestFilters.TAG_TWO],
                         Tags.filter_on_name(
                             [TestFilters.TAG_ONE, TestFilters.TAG_TWO, TestFilters.TAG_THREE],
                             TestFilters.TAG_TWO.name
                         ))

    def test_filter_time_entries_on_project(self) -> None:
        self.assertEqual([TestFilters.TIME_ENTRY_ONE, TestFilters.TIME_ENTRY_TWO],
                         TimeEntries.filter_on_project(
                             [TestFilters.TIME_ENTRY_ONE, TestFilters.TIME_ENTRY_TWO],
                             TestFilters.PROJECT_ONE
                         ))

    def test_filter_time_entries_on_tag_name(self) -> None:
        self.assertEqual([TestFilters.TIME_ENTRY_TWO],
                         TimeEntries.filter_on_any_tags(
                             [TestFilters.TIME_ENTRY_ONE, TestFilters.TIME_ENTRY_TWO],
                             [TestFilters.TAG_THREE]
                         ))


if __name__ == '__main__':
    unittest.main()



from json import JSONDecoder

from togglcmder.toggl.builders.user_builder import UserBuilder


class UserDecoder(JSONDecoder):
    def __init__(self, *args: tuple, **kwargs: dict):
        JSONDecoder.__init__(self,
                             object_hook=UserDecoder.object_hook,
                             *args,
                             **kwargs)

    @staticmethod
    def object_hook(obj: dict):
        if 'data' in obj:
            return obj['data']

        if 'fullname' in obj:
            return UserBuilder()\
                .identifier(obj['id'])\
                .name(obj['fullname'])\
                .api_token(obj['api_token'])\
                .last_updated(last_update=obj['at'])\
                .build()

        return obj



from typing import List, Tuple

from datetime import datetime, timedelta

from togglcmder.toggl.types.time_entry import TimeEntry as TimeEntryType
from togglcmder.toggl.types.workspace import Workspace as WorkspaceType
from togglcmder.toggl.types.project import Project as ProjectType


class TimeEntry(object):
    """
    A composition of the data we want to show in a view
    of a given time entry.
    """

    def __init__(self,
                 time_entries: List[TimeEntryType],
                 project: ProjectType,
                 workspace: WorkspaceType):
        self.__time_entries = time_entries
        self.__project = project
        self.__workspace = workspace

    @staticmethod
    def headers() -> Tuple:
        return (
            "Description",
            "Project Name",
            "Workspace Name",
            # "Start Time",
            # "Stop Time",
            "Duration",
            "Tags",
            "Last Updated"
        )

    def values(self) -> List[Tuple]:
        return [(
            time_entry.description,
            self.__project.name if self.__project else "",
            self.__workspace.name,
            # time_entry.start_time.isoformat(),
            # time_entry.stop_time.isoformat() if time_entry.stop_time else "",
            str(timedelta(seconds=time_entry.duration if time_entry.duration > 0 else
                time_entry.start_time.timestamp() - datetime.now().timestamp())),
            ",".join(tag for tag in time_entry.tags) if time_entry.tags else "",
            time_entry.last_updated.isoformat(),
        ) for time_entry in self.__time_entries]

    def values_sorted(self) -> List[Tuple]:
        return sorted(self.values(),
                      key=lambda entry: entry[3])



from json import JSONEncoder

from togglcmder.toggl.types.tag import Tag


class TagEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Tag):
            return {'tag': {
                'name': obj.name,
                'wid': obj.workspace_identifier
            }}
        return super(TagEncoder, self).default(obj)



"""

"""


class Users(object):
    """

    """

    def __init__(self, url: str):
        """

        :param url:
        """
        self.__url = url

    def __repr__(self):
        return self.__url + "/me"

    def details(self, extra_data: bool = False) -> str:
        if extra_data:
            return self.__url + "/me?with_related_data=true"
        return self.__url + "/me"

    def reset_api_token(self):
        return self.__url + "/reset_token"



from __future__ import annotations
from datetime import datetime
from tzlocal import get_localzone
from typing import Optional

from togglcmder.toggl.types.project import Project


class ProjectBuilder(object):
    def __init__(self, project: Optional[Project] = None):
        if project is not None:
            self.__identifier = project.identifier
            self.__name = project.name
            self.__workspace_identifier = project.workspace_identifier
            self.__color = project.color
            self.__last_updated = project.last_updated
            self.__created = project.created
        else:
            self.__identifier = None
            self.__name = None
            self.__workspace_identifier = None
            self.__color = None
            self.__last_updated = None
            self.__created = None

    def identifier(self, identifier: int) -> ProjectBuilder:
        self.__identifier = identifier
        return self

    def name(self, name: str) -> ProjectBuilder:
        self.__name = name
        return self

    def workspace_identifier(self, workspace_identifier: int) -> ProjectBuilder:
        self.__workspace_identifier = workspace_identifier
        return self

    def color(self, color: int) -> ProjectBuilder:
        self.__color = Project.Color(color)
        return self

    def last_updated(self, *, last_update: Optional[str] = None,
                     epoch: Optional[int] = None) -> ProjectBuilder:
        if last_update:
            self.__last_updated = self.__datetime_from_str(last_update)
        elif epoch:
            self.__last_updated = self.__datetime_from_timestamp(epoch)
        return self

    def created(self, *, created: Optional[str] = None,
                epoch: Optional[int] = None) -> ProjectBuilder:
        if created:
            self.__created = self.__datetime_from_str(created)
        elif epoch:
            self.__created = self.__datetime_from_timestamp(epoch)
        return self

    def build(self) -> Project:
        return Project(
            identifier=self.__identifier,
            name=self.__name,
            workspace_identifier=self.__workspace_identifier,
            color=self.__color,
            last_updated=self.__last_updated,
            created=self.__created)

    @staticmethod
    def __datetime_from_str(date: str) -> datetime:
        parsed_date = datetime.fromisoformat(date.replace('Z', '+00:00'))
        if not parsed_date.tzinfo:
            return get_localzone().localize(parsed_date)
        return parsed_date.astimezone(get_localzone())

    @staticmethod
    def __datetime_from_timestamp(timestamp: int) -> datetime:
        return get_localzone().localize(datetime.fromtimestamp(timestamp))



from typing import List
import re as regex
import logging

from togglcmder.toggl.types.project import Project
from togglcmder.toggl.types.workspace import Workspace


class Projects(object):

    @staticmethod
    def filter_single(
            projects: List[Project],
    ) -> Project:
        if not projects:
            raise RuntimeError('no projects passed in when single project expected')
        elif len(projects) > 1:
            raise RuntimeError('multiple projects passed in when single project expected')

        return projects[0]

    @staticmethod
    def filter_on_name(
            projects: List[Project],
            name: str
    ) -> List[Project]:
        logger = logging.getLogger(__name__)
        if not name:
            logger.debug('no name provided for filter_on_name(projects) with input = {}'.format(
                    [project.name for project in projects]
                ))
            return projects

        result = list(filter(
            lambda project:
                regex.match("^{}$".format(name),
                            project.name,
                            regex.IGNORECASE),
            projects
        ))

        if not result:
            logger.warning('no projects found with name = {} and with input = {}'.format(
                name, [project.name for project in projects]
            ))
        return result

    @staticmethod
    def filter_on_workspace(
            projects: List[Project],
            workspace: Workspace
    ) -> List[Project]:
        logger = logging.getLogger(__name__)
        if not workspace:
            logger.debug('no workspace provided for filter_on_workspace(projects) with input = {}'.format(
                [(project.name, project.workspace_identifier) for project in projects]
            ))
            return projects

        result = list(filter(
            lambda project: workspace.identifier == project.workspace_identifier,
            projects
        ))

        if not result:
            logger.warning('no projects found with workspace = {} and with input = {}'.format(
                (workspace.name, workspace.identifier),
                [(project.name, project.workspace_identifier) for project in projects]
            ))
        return result

    @staticmethod
    def filter_on_color(
            projects: List[Project],
            color: Project.Color
    ) -> List[Project]:
        logger = logging.getLogger(__name__)
        if not color:
            logger.debug('no color provided for filter_on_color(projects) with input = {}'.format(
                [(project.name, project.color.name) for project in projects]
            ))
            return projects

        result = list(filter(
            lambda project: color == project.color,
            projects
        ))

        if not result:
            logger.warning('no projects found with color = {} and with input = {}'.format(
                color.name, [(project.name, project.color.name) for project in projects]
            ))
        return result

    @staticmethod
    def filter_on_identifier(
            projects: List[Project],
            identifier: int
    ) -> List[Project]:
        logger = logging.getLogger(__name__)
        if not identifier:
            logger.debug('no identifier provided for filter_on_identifier(projects) with input = {}'.format(
                [(project.name, project.identifier) for project in projects]
            ))
            return projects

        result = list(filter(
            lambda project: identifier == project.identifier,
            projects
        ))

        if not result:
            logger.warning('no projects found with identifier = {} and with input = {}'.format(
                identifier, [(project.name, project.identifier) for project in projects]
            ))
        return result



"""

"""
from datetime import datetime
from typing import List, Optional

import urllib.parse


class TimeEntries(object):
    """

    """

    def __init__(self, url: str):
        self.__url = url + "/time_entries"

    def __repr__(self):
        return self.__url

    def start(self) -> str:
        """
        Used to start a new entry, with details in a POST request.

        :return:
        """
        return self.__url + "/start"

    def stop(self, id_: int) -> str:
        """
        Used to stop a specific entry.

        :param id_:
        :return:
        """
        return self.__url + "/{}/stop".format(id_)

    def current(self) -> str:
        """
        Only used to return the current running entry.

        :return:
        """
        return self.__url + "/current"

    def details(self, identifier: Optional[int] = None, identifiers: Optional[List[int]] = None) -> str:
        """
        This can be used for multiple types of requests, such as:
            1. Simply getting more details on the entry.
            2. Updating the entry.
            3. Deleting the entry.
            4. Handling more than one time entry.

        :param identifier:
        :param identifiers:
        :return:
        """
        if identifier:
            return self.__url + "/{}".format(identifier)
        elif identifiers:
            return self.__url + "/{}".format(",".join(map(str, identifiers)))
        else:
            return self.__url

    # Used to search for time entries with a GET request.
    def search(self, start: datetime, end: datetime) -> str:
        return self.__url + "?start_date={}&end_date={}".format(
            urllib.parse.quote(start.isoformat()),
            urllib.parse.quote(end.isoformat()))



from typing import List
import re as regex
import logging

from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.types.tag import Tag


class Tags(object):

    @staticmethod
    def filter_single(
            tags: List[Tag],
    ) -> Tag:
        if not tags:
            raise RuntimeError('no tags passed in when single tag expected')
        elif len(tags) > 1:
            raise RuntimeError('multiple tags passed in when single tag expected')

        return tags[0]

    @staticmethod
    def filter_on_name(
            tags: List[Tag],
            name: str
    ) -> List[Tag]:
        logger = logging.getLogger(__name__)
        if not name:
            logger.debug('no name provided for filter_on_name(tags) with input = {}'.format(
                [tag.name for tag in tags]
            ))
            return tags

        result = list(filter(
            lambda tag:
                regex.match("^{}$".format(name),
                            tag.name,
                            regex.IGNORECASE),
            tags
        ))

        if not result:
            logger.warning('no tags found with name = {} and with input = {}'.format(
                name, [tag.name for tag in tags]
            ))
        return result

    @staticmethod
    def filter_on_names(
            tags: List[Tag],
            names: List[str]
    ) -> List[Tag]:
        logger = logging.getLogger(__name__)
        if not names:
            logger.debug('no names provided for filter_on_names(tags) with input = {}'.format(
                [tag.name for tag in tags]
            ))
            return tags

        result = list(filter(
            lambda tag: tag.name in names,
            tags
        ))

        if not result:
            logger.warning('no tags found with names = {} and with input = {}'.format(
                names, [tag.name for tag in tags]
            ))
        return result

    @staticmethod
    def filter_on_workspace(
            tags: List[Tag],
            workspace: Workspace
    ) -> List[Tag]:
        logger = logging.getLogger(__name__)
        if not workspace:
            logger.debug('no workspace provided for filter_on_workspace(tags) with input = {}'.format(
                [(tag.name, tag.workspace_identifier) for tag in tags]
            ))
            return tags

        result = list(filter(
            lambda tag: workspace.identifier == tag.workspace_identifier,
            tags
        ))

        if not result:
            logger.warning('no tags found with workspace = {} and with input = {}'.format(
                (workspace.name, workspace.identifier), [(tag.name, tag.workspace_identifier) for tag in tags]
            ))
        return result

    @staticmethod
    def filter_on_identifier(
            tags: List[Tag],
            identifier: int
    ) -> List[Tag]:
        logger = logging.getLogger(__name__)
        if not identifier:
            logger.debug('no identifier provided for filter_on_identifier(tags) with input = {}'.format(
                [(tag.name, tag.identifier) for tag in tags]
            ))
            return tags

        result = list(filter(
            lambda tag: identifier == tag.identifier,
            tags
        ))

        if not result:
            logger.warning('no tags found with identifier = {} and with input = {}'.format(
                identifier, [(tag.name, tag.identifier) for tag in tags]
            ))
        return result



from __future__ import annotations

from datetime import datetime
from pytz import timezone
from tzlocal import get_localzone
from typing import List, Optional

from togglcmder.toggl.types.time_entry import TimeEntry


class TimeEntryBuilder(object):
    def __init__(self, time_entry: Optional[TimeEntry] = None):
        if time_entry is not None:
            self.__identifier = time_entry.identifier
            self.__description = time_entry.description
            self.__project_identifier = time_entry.project_identifier
            self.__workspace_identifier = time_entry.workspace_identifier
            self.__start_time = time_entry.start_time
            self.__duration = time_entry.duration
            self.__stop_time = time_entry.stop_time
            self.__tags = time_entry.tags
            self.__last_updated = time_entry.last_updated
        else:
            self.__identifier = None
            self.__description = None
            self.__project_identifier = None
            self.__workspace_identifier = None
            self.__start_time = None
            self.__duration = None
            self.__stop_time = None
            self.__tags = None
            self.__last_updated = None

    def identifier(self, identifier: int) -> TimeEntryBuilder:
        self.__identifier = identifier
        return self

    def description(self, description: str) -> TimeEntryBuilder:
        self.__description = description
        return self

    def project_identifier(self, project_identifier: int) -> TimeEntryBuilder:
        self.__project_identifier = project_identifier
        return self

    def workspace_identifier(self, workspace_identifier: int) -> TimeEntryBuilder:
        self.__workspace_identifier = workspace_identifier
        return self

    def start_time(self, *, start_time: Optional[str] = None, epoch: Optional[int] = None,
                   dt: Optional[datetime] = None) -> TimeEntryBuilder:
        if start_time is not None:
            self.__start_time = self.__datetime_from_str(start_time)
        elif epoch is not None:
            self.__start_time = self.__datetime_from_timestamp(epoch)
        elif dt is not None:
            self.__start_time = dt
        return self

    def duration(self, duration: int) -> TimeEntryBuilder:
        if duration < 0:
            self.__duration = datetime.now(timezone('UTC')).replace(microsecond=0).timestamp() + duration
        else:
            self.__duration = duration
        return self

    def stop_time(self, *, stop_time: Optional[str] = None, epoch: Optional[int] = None,
                  dt: Optional[datetime] = None) -> TimeEntryBuilder:
        if stop_time is not None:
            self.__stop_time = self.__datetime_from_str(stop_time)
        elif epoch is not None:
            self.__stop_time = self.__datetime_from_timestamp(epoch)
        elif dt is not None:
            self.__stop_time = dt
        return self

    def unset_stop_time(self):
        self.__stop_time = None

    def unset_duration(self):
        self.__duration = None

    def tags(self, tags: List[str]) -> TimeEntryBuilder:
        self.__tags = tags
        return self

    def add_tags(self, tags: List[str]) -> TimeEntryBuilder:
        self.__tags.extend(tags)
        return self

    def remove_tags(self, tags: List[str]) -> TimeEntryBuilder:
        for tag in tags:
            self.__tags.remove(tag)
        return self

    def last_updated(self, *, last_update: Optional[str] = None, epoch: Optional[int] = None) -> TimeEntryBuilder:
        if last_update is not None:
            self.__last_updated = self.__datetime_from_str(last_update)
        elif epoch is not None:
            self.__last_updated = self.__datetime_from_timestamp(epoch)
        return self

    def build(self) -> TimeEntry:
        return TimeEntry(
            identifier=self.__identifier,
            description=self.__description,
            project_identifier=self.__project_identifier,
            workspace_identifier=self.__workspace_identifier,
            start_time=self.__start_time,
            duration=self.__duration,
            stop_time=self.__stop_time,
            tags=self.__tags,
            last_updated=self.__last_updated)

    @staticmethod
    def __datetime_from_str(date: str) -> datetime:
        parsed_date = datetime.fromisoformat(date.replace('Z', '+00:00'))
        if not parsed_date.tzinfo:
            return get_localzone().localize(parsed_date)
        return parsed_date.astimezone(tz=get_localzone())

    @staticmethod
    def __datetime_from_timestamp(timestamp: int) -> datetime:
        return get_localzone().localize(datetime.fromtimestamp(timestamp))



from typing import Optional
from functools import total_ordering


@total_ordering
class Tag(object):
    def __init__(self, *, name: str, workspace_identifier: int,
                 identifier: Optional[int] = None):
        self.__name = name
        self.__workspace_identifier = workspace_identifier
        self.__identifier = identifier

    def __eq__(self, other):
        if not isinstance(other, Tag):
            return NotImplemented

        return self.__name == other.name \
            and self.__workspace_identifier == other.workspace_identifier \
            and self.__identifier == other.identifier

    def __lt__(self, other):
        if not isinstance(other, Tag):
            raise NotImplemented
        return self.__name.lower() < other.__name.lower()

    def __str__(self):
        return "{},{},{}".format(
            self.name,
            self.identifier,
            self.workspace_identifier)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def identifier(self) -> int:
        return self.__identifier

    @property
    def workspace_identifier(self) -> int:
        return self.__workspace_identifier



from setuptools import setup, find_packages

from togglcmder.version import __version__

with open("README.md", "r") as readme:
    long_description = readme.read()

with open("requirements.txt", "r") as requirements:
    required_packages = requirements.read()

setup(
    name='togglcmder',
    version=__version__,
    url="https://github.com/yatesjr/toggl-cmder",
    author="Jonathan Yates",
    author_email="yatesjr@avernakis.com",
    description="Utility to control Toggl timers via the REST API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    packages=find_packages(),
    install_requires=required_packages.split(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    entry_points={
      'console_scripts': [
          'togglcmder = togglcmder.__main__:main'
      ]
    },
    python_requires='>=3.6'
)



from typing import List, Tuple

from togglcmder.toggl.types.workspace import Workspace as WorkspaceType


class Workspace(object):
    """
    A composition of the data we want to show in a view
    of a given workspace.
    """

    def __init__(self, workspaces: List[WorkspaceType]):
        self.__workspaces = workspaces

    @staticmethod
    def headers() -> Tuple:
        return (
            "Identifier",
            "Workspace Name",
            "Last Updated"
        )

    def values(self) -> List[Tuple]:
        return [(
            workspace.identifier,
            workspace.name,
            workspace.last_updated.isoformat()
        ) for workspace in self.__workspaces]



import unittest
from datetime import datetime
from tzlocal import get_localzone

from togglcmder.toggl.caching import Caching
from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.types.project import Project
from togglcmder.toggl.types.tag import Tag
from togglcmder.toggl.types.user import User
from togglcmder.toggl.types.time_entry import TimeEntry


class TestCaching(unittest.TestCase):
    WORKSPACE = Workspace(
        name='Test Workspace',
        identifier=1,
        last_updated=get_localzone().localize(datetime.now()))

    PROJECT = Project(
        name='Test Project',
        color=Project.Color.RED,
        last_updated=get_localzone().localize(datetime.now()),
        created=get_localzone().localize(datetime.now()),
        identifier=1,
        workspace_identifier=1)

    TAG_ONE = Tag(
        name='Test Tag One',
        identifier=1,
        workspace_identifier=1)

    TAG_TWO = Tag(
        name='Test Tag Two',
        identifier=2,
        workspace_identifier=1
    )

    TAG_THREE = Tag(
        name='Test Tag Three',
        identifier=3,
        workspace_identifier=1
    )

    TIME_ENTRY_ONE = TimeEntry(
        description='Test Entry One',
        start_time=get_localzone().localize(datetime.now()),
        identifier=1,
        duration=30,
        project_identifier=1,
        workspace_identifier=1,
        stop_time=get_localzone().localize(datetime.now()),
        tags=['Test Tag One', 'Test Tag Two'],
        last_updated=get_localzone().localize(datetime.now())
    )

    TIME_ENTRY_TWO = TimeEntry(
        description='Test Entry Two',
        start_time=get_localzone().localize(datetime.now()),
        identifier=2,
        duration=60,
        project_identifier=1,
        workspace_identifier=1,
        stop_time=get_localzone().localize(datetime.now()),
        tags=['Test Tag Three'],
        last_updated=get_localzone().localize(datetime.now())
    )

    USER = User(
        name='Test User',
        api_token='1234',
        identifier=1,
        last_updated=get_localzone().localize(datetime.now()))

    def setUp(self) -> None:
        self.__connection = Caching(cache_name=':memory:')

    def tearDown(self) -> None:
        del self.__connection

    def test_workspace_caching(self):
        self.assertEqual(1, self.__connection.update_workspace_cache([TestCaching.WORKSPACE]))
        print(self.__connection.retrieve_workspace_cache())
        self.assertEqual([TestCaching.WORKSPACE], self.__connection.retrieve_workspace_cache())

    def test_project_caching(self):
        self.assertEqual(1, self.__connection.update_workspace_cache([TestCaching.WORKSPACE]))
        self.assertEqual(1, self.__connection.update_project_cache([TestCaching.PROJECT]))
        self.assertEqual([TestCaching.PROJECT], self.__connection.retrieve_project_cache())

    def test_tag_caching(self):
        self.assertEqual(1, self.__connection.update_workspace_cache([TestCaching.WORKSPACE]))
        self.assertEqual(3, self.__connection.update_tag_cache([TestCaching.TAG_ONE,
                                                             TestCaching.TAG_TWO,
                                                             TestCaching.TAG_THREE]))
        self.assertEqual([TestCaching.TAG_ONE,
                          TestCaching.TAG_TWO,
                          TestCaching.TAG_THREE], self.__connection.retrieve_tag_cache())

    def test_time_entry_caching(self):
        self.assertEqual(1, self.__connection.update_workspace_cache([TestCaching.WORKSPACE]))
        self.assertEqual(1, self.__connection.update_project_cache([TestCaching.PROJECT]))
        self.assertEqual(3, self.__connection.update_tag_cache(
            [TestCaching.TAG_ONE, TestCaching.TAG_TWO, TestCaching.TAG_THREE]))
        self.assertEqual(2, self.__connection.update_time_entry_cache(
            [TestCaching.TIME_ENTRY_ONE, TestCaching.TIME_ENTRY_TWO]))
        self.assertEqual([TestCaching.TIME_ENTRY_ONE, TestCaching.TIME_ENTRY_TWO],
                         self.__connection.retrieve_time_entry_cache())

    def test_user_caching(self):
        self.assertEqual(1, self.__connection.update_workspace_cache([TestCaching.WORKSPACE]))
        self.assertEqual(1, self.__connection.update_user_cache(TestCaching.USER))
        self.assertEqual(TestCaching.USER, self.__connection.retrieve_user_cache())


if __name__ == '__main__':
    unittest.main()



import sqlite3
from sqlite3 import IntegrityError
import logging

from typing import List
from datetime import datetime

from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.builders.workspace_builder import WorkspaceBuilder
from togglcmder.toggl.types.time_entry import TimeEntry
from togglcmder.toggl.builders.time_entry_builder import TimeEntryBuilder
from togglcmder.toggl.types.user import User
from togglcmder.toggl.builders.user_builder import UserBuilder
from togglcmder.toggl.types.tag import Tag
from togglcmder.toggl.builders.tag_builder import TagBuilder
from togglcmder.toggl.types.project import Project
from togglcmder.toggl.builders.project_builder import ProjectBuilder


class Caching(object):
    WORKSPACE_TABLE = '''
    CREATE TABLE IF NOT EXISTS workspaces (
        name TEXT NOT NULL,
        identifier INTEGER PRIMARY KEY,
        last_updated TIMESTAMP NOT NULL
    )
    '''

    PROJECT_TABLE = '''
    CREATE TABLE IF NOT EXISTS projects (
        name TEXT NOT NULL,
        color INTEGER,
        last_updated TIMESTAMP NOT NULL,
        created TIMESTAMP NOT NULL,
        identifier INTEGER PRIMARY KEY,
        workspace_identifier INTEGER NOT NULL,
        FOREIGN KEY (workspace_identifier) REFERENCES workspaces (identifier) ON DELETE CASCADE
    )
    '''

    TAG_TABLE = '''
    CREATE TABLE IF NOT EXISTS tags (
        name TEXT,
        identifier INTEGER PRIMARY KEY,
        workspace_identifier INTEGER,
        FOREIGN KEY (workspace_identifier) REFERENCES workspaces (identifier) ON DELETE CASCADE
    )
    '''

    TIME_ENTRY_TABLE = '''
    CREATE TABLE IF NOT EXISTS time_entries (
        description TEXT,
        start_time TIMESTAMP NOT NULL,
        stop_time TIMESTAMP,
        duration INTEGER,
        identifier INTEGER PRIMARY KEY,
        project_identifier INTEGER,
        workspace_identifier INTEGER NOT NULL,
        last_updated TIMESTAMP,
        FOREIGN KEY (project_identifier) REFERENCES projects (identifier) ON DELETE CASCADE,
        FOREIGN KEY (workspace_identifier) REFERENCES workspaces (identifier)
    )
    '''

    TIME_ENTRY_TAG_JUNCTION_TABLE = '''
    CREATE TABLE IF NOT EXISTS time_entry_tags (
        tag_identifier INTEGER NOT NULL,
        time_entry_identifier INTEGER NOT NULL,
        FOREIGN KEY (tag_identifier) REFERENCES tags (identifier) ON DELETE CASCADE,
        FOREIGN KEY (time_entry_identifier) REFERENCES time_entries (identifier) ON DELETE CASCADE 
    )
    '''

    USER_TABLE = '''
    CREATE TABLE IF NOT EXISTS users (
        name TEXT,
        api_token TEXT,
        identifier INTEGER PRIMARY KEY,
        last_updated TIMESTAMP NOT NULL
    )
    '''

    def __init__(self, *, cache_name: str = "cache.db"):
        self.__connection = sqlite3.connect(cache_name)
        self.__connection.set_trace_callback(logging.getLogger(__name__).debug)
        self.__cursor = self.__connection.cursor()

        self.__cursor.execute("PRAGMA foreign_keys = 1")
        self.__connection.commit()

        self.__cursor.execute(Caching.WORKSPACE_TABLE)
        self.__cursor.execute(Caching.PROJECT_TABLE)
        self.__cursor.execute(Caching.TAG_TABLE)
        self.__cursor.execute(Caching.TIME_ENTRY_TABLE)
        self.__cursor.execute(Caching.TIME_ENTRY_TAG_JUNCTION_TABLE)
        self.__cursor.execute(Caching.USER_TABLE)

        self.__connection.commit()

        self.__workspaces: List[Workspace] = []
        self.__projects: List[Project] = []
        self.__tags: List[Tag] = []
        self.__time_entries: List[TimeEntry] = []

    def __del__(self):
        self.__connection.close()

    def update_workspace_cache(self, workspaces: List[Workspace]) -> int:
        insert_sql = '''
            INSERT INTO workspaces
            (name, identifier, last_updated) VALUES 
            (?, ?, ?)
        '''

        update_sql = '''
            UPDATE workspaces SET name=?, last_updated=?
            WHERE identifier=?
        '''

        for workspace in workspaces:
            try:
                self.__cursor.execute(
                    insert_sql, (workspace.name,
                                 workspace.identifier,
                                 workspace.last_updated.timestamp()))
            except IntegrityError:
                self.__cursor.execute(
                    update_sql, (workspace.name,
                                 workspace.last_updated.timestamp(),
                                 workspace.identifier))

        self.__connection.commit()
        return self.__cursor.rowcount

    def retrieve_workspace_cache(self) -> List[Workspace]:
        sql = '''
            SELECT name, identifier, last_updated FROM workspaces
        '''

        self.__cursor.execute(sql)
        results = self.__cursor.fetchall()
        if results:
            return [
                WorkspaceBuilder()
                    .name(result[0])
                    .identifier(result[1])
                    .last_updated(epoch=result[2]).build()
                for result in results
            ]

    def update_user_cache(self, user: User) -> int:
        insert_sql = '''
            INSERT INTO users
            (name, api_token, identifier, last_updated) VALUES 
            (?, ?, ?, ?)
        '''

        update_sql = '''
            UPDATE users SET name=?, api_token=?, last_updated=?
            WHERE identifier=?
        '''

        try:
            self.__cursor.execute(
                insert_sql, (user.name,
                             user.api_token,
                             user.identifier,
                             user.last_updated.timestamp()))
        except IntegrityError:
            self.__cursor.execute(
                update_sql, (user.name,
                             user.api_token,
                             user.last_updated.timestamp(),
                             user.identifier))

        self.__connection.commit()
        return self.__cursor.rowcount

    def retrieve_user_cache(self) -> User:
        sql = '''
            SELECT name, api_token, identifier, last_updated FROM users
        '''

        self.__cursor.execute(sql)
        results = self.__cursor.fetchone()
        if results:
            return UserBuilder()\
                .name(results[0])\
                .api_token(results[1])\
                .identifier(results[2])\
                .last_updated(epoch=results[3]).build()

    def update_project_cache(self, projects: List[Project]) -> int:
        insert_sql = '''
            INSERT INTO projects
            (name, color, last_updated, created, identifier, workspace_identifier) VALUES 
            (?, ?, ?, ?, ?, ?) 
        '''

        update_sql = '''
            UPDATE projects
            SET name=?, color=?, last_updated=?, workspace_identifier=?
            WHERE identifier=?
        '''

        for project in projects:
            try:
                self.__cursor.execute(
                    insert_sql, (project.name,
                                 project.color.value,
                                 project.last_updated.timestamp(),
                                 project.created.timestamp() if project.created else datetime.now().timestamp(),
                                 project.identifier,
                                 project.workspace_identifier))
            except IntegrityError:
                self.__cursor.execute(
                    update_sql, (project.name,
                                 project.color.value,
                                 project.last_updated.timestamp(),
                                 project.workspace_identifier,
                                 project.identifier))

        self.__connection.commit()
        return self.__cursor.rowcount

    def retrieve_project_cache(self) -> List[Project]:
        sql = '''
            SELECT name, color, last_updated, created, identifier, workspace_identifier FROM projects
        '''

        self.__cursor.execute(sql)
        results = self.__cursor.fetchall()
        if results:
            return [
                ProjectBuilder()
                    .name(result[0])
                    .color(result[1])
                    .last_updated(epoch=result[2])
                    .created(epoch=result[3])
                    .identifier(result[4])
                    .workspace_identifier(result[5]).build()
                for result in results
            ]

    def remove_project_from_cache(self, project: Project) -> None:
        sql = '''
            DELETE FROM projects
            WHERE identifier=?
        '''
        self.__cursor.execute(sql, (project.identifier,))

        self.__connection.commit()

    def update_tag_cache(self, tags: List[Tag]) -> int:
        insert_sql = '''
            INSERT INTO tags
            (name, identifier, workspace_identifier) VALUES
            (?, ?, ?)
        '''

        update_sql = '''
            UPDATE tags
            SET name=?, workspace_identifier=?
            WHERE identifier=?
        '''

        rows_affected = 0
        for tag in tags:
            try:
                self.__cursor.execute(
                    insert_sql, (tag.name,
                                 tag.identifier,
                                 tag.workspace_identifier))
            except IntegrityError:
                self.__cursor.execute(
                    update_sql, (tag.name,
                                 tag.workspace_identifier,
                                 tag.identifier))
            rows_affected += self.__cursor.rowcount

        self.__connection.commit()
        return rows_affected

    def retrieve_tag_cache(self) -> List[Tag]:
        sql = """
            SELECT name, identifier, workspace_identifier FROM tags
        """

        self.__cursor.execute(sql)
        results = self.__cursor.fetchall()
        if results:
            return [
                TagBuilder()
                    .name(result[0])
                    .identifier(result[1])
                    .workspace_identifier(result[2]).build()
                for result in results
            ]

    def remove_tag_from_cache(self, tag: Tag) -> None:
        tag_removal_sql = '''
            DELETE FROM tags
            WHERE identifier=?
        '''
        self.__cursor.execute(tag_removal_sql, (tag.identifier,))

        join_table_removal_sql = '''
            DELETE FROM time_entry_tags
            WHERE tag_identifier=?
        '''
        self.__cursor.execute(join_table_removal_sql, (tag.identifier,))

        self.__connection.commit()

    def __retrieve_time_entry_tags_join(self, time_entry_identifier: int) -> List[tuple]:
        sql = '''
            SELECT name, tag_identifier, time_entry_identifier
            FROM time_entry_tags
            INNER JOIN tags ON time_entry_tags.tag_identifier = tags.identifier
            WHERE time_entry_identifier=?
        '''
        self.__cursor.execute(sql, (time_entry_identifier,))
        return self.__cursor.fetchall()

    def __retrieve_time_entry_tags(self, time_entry_identifier: int) -> List[tuple]:
        sql = '''
            SELECT tag_identifier, time_entry_identifier
            FROM time_entry_tags
            WHERE time_entry_identifier=?
        '''
        self.__cursor.execute(sql, (time_entry_identifier,))
        return self.__cursor.fetchall()

    def __check_existing(self, tags: List[int], time_entry_identifier: int) -> List[int]:
        # returns a tuple of (tag_id, time_entry_id)
        existing_time_entry_tags = self.__retrieve_time_entry_tags(time_entry_identifier)
        if len(tags) == 0 or len(existing_time_entry_tags) == 0:
            return tags
        return list(
            # 2. map each tuple to be just the tag identifier, so given x,
            #    return the tag_id
            map(lambda x: x[0],
                # 1. filter so we only get tags tuples that aren't in the existing list
                #     of tags (checked based on x[0], which is the tag identifier)
                filter(lambda x: x[0] not in tags, existing_time_entry_tags)))

    def update_time_entry_cache(self, time_entries: List[TimeEntry]) -> int:
        insert_sql = '''
            INSERT INTO time_entries
            (description, start_time, stop_time, duration, identifier,
             project_identifier, workspace_identifier, last_updated) VALUES 
            (?, ?, ?, ?, ?, ?, ?, ?)
        '''

        update_sql = '''
            UPDATE time_entries
            SET description=?, 
                start_time=?, 
                stop_time=?, 
                duration=?, 
                project_identifier=?, 
                workspace_identifier=?, 
                last_updated=?
            WHERE identifier=?
        '''

        insert_time_entry_tag_sql = '''
            INSERT INTO time_entry_tags
            (tag_identifier, time_entry_identifier)
            VALUES (?, ?)
        '''
        tag_rows = self.retrieve_tag_cache()
        rows_affected = 0

        for time_entry in time_entries:
            try:
                self.__cursor.execute(
                    insert_sql, (time_entry.description,
                                 time_entry.start_time.timestamp(),
                                 None if not time_entry.stop_time else time_entry.stop_time.timestamp(),
                                 time_entry.duration,
                                 time_entry.identifier,
                                 time_entry.project_identifier,
                                 time_entry.workspace_identifier,
                                 time_entry.last_updated.timestamp()))
            except IntegrityError:
                self.__cursor.execute(
                    update_sql, (time_entry.description,
                                 time_entry.start_time.timestamp(),
                                 None if not time_entry.stop_time else time_entry.stop_time.timestamp(),
                                 time_entry.duration,
                                 time_entry.project_identifier,
                                 time_entry.workspace_identifier,
                                 time_entry.last_updated.timestamp(),
                                 time_entry.identifier))
            rows_affected += self.__cursor.rowcount

            tag_ids = []
            if time_entry.tags:
                for tag in time_entry.tags:
                    for tag_row in tag_rows:
                        if tag == tag_row.name:
                            tag_ids.append(tag_row.identifier)
                            break
                for tag_id in self.__check_existing(tag_ids, time_entry.identifier):
                    self.__cursor.execute(
                        insert_time_entry_tag_sql,
                        (tag_id,
                         time_entry.identifier))

        self.__connection.commit()

        return rows_affected

    def retrieve_time_entry_cache(self) -> List[TimeEntry]:
        time_entry_sql = """
            SELECT  description, 
                    start_time, 
                    stop_time, 
                    duration, 
                    identifier, 
                    project_identifier, 
                    workspace_identifier, 
                    last_updated
            FROM time_entries
        """

        time_entries = []

        self.__cursor.execute(time_entry_sql)
        results = self.__cursor.fetchall()
        for result in results:
            tag_results = self.__retrieve_time_entry_tags_join(result[4])

            builder = TimeEntryBuilder()\
                .description(result[0])\
                .start_time(epoch=result[1])\
                .stop_time(epoch=result[2])\
                .duration(result[3])\
                .identifier(result[4])\
                .project_identifier(result[5])\
                .workspace_identifier(result[6])\
                .last_updated(epoch=result[7])\
                .tags([tag_result[0] for tag_result in tag_results])
            time_entries.append(builder.build())

        return time_entries

    def remove_time_entry_from_cache(self, time_entry: TimeEntry) -> None:
        entry_removal_sql = '''
            DELETE FROM time_entries
            WHERE identifier=?
        '''
        self.__cursor.execute(entry_removal_sql, (time_entry.identifier,))

        joined_entry_removal_sql = '''
            DELETE FROM time_entry_tags
            WHERE time_entry_identifier=?
        '''
        self.__cursor.execute(joined_entry_removal_sql, (time_entry.identifier,))

        self.__connection.commit()

    def get_workspace_identifier(self, workspace_name: str) -> int:
        sql = """
            SELECT identifier
            FROM workspaces
            WHERE name=?
        """
        self.__cursor.execute(sql, (workspace_name,))
        return self.__cursor.fetchone()[0]

    def get_project_identifier(self, project_name: str) -> int:
        sql = """
            SELECT identifier
            FROM projects
            WHERE name=?
        """
        self.__cursor.execute(sql, (project_name,))
        return self.__cursor.fetchone()[0]



import click
from tabulate import tabulate
import logging
from typing import List, Optional

from togglcmder.toggl.cli.workspaces \
    import sync_or_retrieve_workspaces, retrieve_workspace_from_context

from togglcmder.toggl.cli.helpers import retrieve_cache_from_context
from togglcmder.toggl.cli.helpers import retrieve_commands_from_context
from togglcmder.toggl.cli.helpers import retrieve_downloader_from_context

from togglcmder.toggl.types.workspace import Workspace
from togglcmder.toggl.filters.workspaces import Workspaces as WorkspaceFilter

from togglcmder.toggl.types.project import Project
from togglcmder.toggl.builders.project_builder import ProjectBuilder
from togglcmder.toggl.views.project import Project as ProjectView
from togglcmder.toggl.filters.projects import Projects as ProjectFilter


def sync_or_retrieve_projects(context_obj: dict, workspace: Workspace) -> List[Project]:
    caching = retrieve_cache_from_context(context_obj)
    downloader = retrieve_downloader_from_context(context_obj)

    # Download any new projects if remote sync is enabled.
    if context_obj['sync'] is True:
        current_projects = downloader.download_projects(workspace)
        if current_projects:
            caching.update_project_cache(current_projects)
    else:
        # Otherwise we just pull from the local cache.
        current_projects = caching.retrieve_project_cache()

        if current_projects:
            # Retrieval from DB doesn't filter on workspace, so we do that ourselves.
            current_projects = ProjectFilter.filter_on_workspace(current_projects, workspace)

    return current_projects


def retrieve_project_from_context(context: click.Context) -> Optional[Project]:
    project = context.obj['data']['project']
    return project


@click.group(
    help='Add, update, delete, and list projects.'
)
@click.option('--workspace')
@click.pass_context
def projects(context: click.Context, workspace: str):
    workspace_name = workspace or context.obj['config']['default_workspace']

    workspaces = sync_or_retrieve_workspaces(context.obj)

    if workspace_name:
        workspaces = WorkspaceFilter.filter_on_name(
            workspaces,
            workspace_name)

        workspaces_found = len(workspaces)
        if workspaces_found == 1:
            context.obj['data']['workspace'] = workspaces[0]

    context.obj['data']['workspaces'] = workspaces

    current_projects = []
    for current_workspace in workspaces:
        synced_projects = sync_or_retrieve_projects(context.obj, current_workspace)
        if synced_projects:
            current_projects.extend(synced_projects)

    context.obj['data']['projects'] = current_projects


@projects.command('add')
@click.option('--name',
              required=True)
@click.option('--color',
              type=click.Choice([val.__str__() for val in Project.Color]),
              default=Project.Color.BLACK.name.lower())
@click.pass_context
def project_add(context: click.Context, name: str, color: str):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when adding a new project.')
        return

    # If there is a project with the same name existing already, do nothing.
    current_projects = context.obj['data']['projects']
    if current_projects:
        current_projects = ProjectFilter.filter_on_name(
            current_projects, name)
        if current_projects:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': project({name}) with this name already'
                       f' exists in this workspace({workspace.name}).')
            return

    project = ProjectBuilder() \
        .name(name) \
        .workspace_identifier(workspace.identifier) \
        .color(Project.Color.from_string(color)) \
        .build()

    commands = retrieve_commands_from_context(context.obj)
    cache = retrieve_cache_from_context(context.obj)

    # This raises if there is any issues, which means it doesn't get added
    # to the local cache.
    try:
        added_project = commands.add_project(project)
        # The new project was added to the remote Toggl servers, so now
        # we can add it to our local cache as well.
        cache.update_project_cache([added_project])

        click.echo(click.style('SUCCESS', fg='green') +
                   f': added new project({added_project.name}) to'
                   f' workspace({workspace.name})!')

    except Exception as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to add new project({name}) to workspace({workspace.name}).'
                   ' An exception has been logged; check the logs for more information.')


@projects.command('delete')
@click.option('--name')
@click.option('--color',
              type=click.Choice([val.__str__() for val in Project.Color]))
@click.option('--multiple',
              is_flag=True,
              default=False,
              show_default=True)
@click.pass_context
def project_delete(context: click.Context, name: str, color: str,
                   multiple: bool):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when deleting a project.')
        return

    current_projects = context.obj['data']['projects']
    if not current_projects:
        click.echo(click.style('WARNING', fg='yellow') +
                   f': no projects exist in this workspace({workspace.name})!')
        return

    if color:
        # See if the user specified a color to search on.
        current_projects = ProjectFilter.filter_on_color(
            current_projects, Project.Color.from_string(color))
        if not current_projects:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no projects exist with this color({color})'
                       f' in this workspace({workspace.name}).')
            return

    if name:
        current_projects = ProjectFilter.filter_on_name(
            current_projects, name)
        if not current_projects:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no projects exist with this name({name}'
                       f' in this workspace({workspace.name}).')
            return

    if not multiple and len(current_projects) > 1:
        # If multiple matches are found but the user didn't specify the --multiple option,
        # prevent accidental deletion by requiring the user do it again but with the option
        # enabled.
        click.echo(click.style('ERROR', fg='red') +
                   ': multiple projects found matching the criteria, but --multiple '
                   'not specified.. please use the --multiple option or tighten '
                   'the search criteria.')
        return

    commands = retrieve_commands_from_context(context.obj)
    cache = retrieve_cache_from_context(context.obj)

    try:
        if len(current_projects) > 1:
            commands.delete_projects(current_projects)
            for project in current_projects:
                cache.remove_project_from_cache(project)
                click.echo(click.style('SUCCESS', fg='green') +
                           f': deleted project({project.name})!')
        else:
            commands.delete_project(current_projects[0])
            cache.remove_project_from_cache(current_projects[0])
            click.echo(click.style('SUCCESS', fg='green') +
                       f': deleted project({current_projects[0].name})!')

    except Exception as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to delete the project(s). An exception has been logged;'
                   ' check the logs for more information.')


@projects.command('update')
@click.option('--old-name')
@click.option('--old-color',
              type=click.Choice([val.__str__() for val in Project.Color]))
@click.option('--new-name')
@click.option('--new-color',
              type=click.Choice([val.__str__() for val in Project.Color]))
@click.option('--multiple',
              is_flag=True,
              default=False,
              show_default=True)
@click.pass_context
def project_update(context: click.Context, old_name: str, old_color: str,
                   new_name: str, new_color: str, multiple: bool):
    workspace = retrieve_workspace_from_context(context)
    if not workspace:
        # Workspace filter was not strict enough or there is no default
        # workspace configured.
        click.echo(click.style('ERROR', fg='red') +
                   ': a single workspace must be specified (default or otherwise'
                   ' ) when updating a project.')
        return

    current_projects = context.obj['data']['projects']
    if not current_projects:
        click.echo(click.style('WARNING', fg='yellow') +
                   f': no projects to update in this workspace({workspace.name}).')
        return

    if old_color:
        # See if the user specified a color to search on.
        current_projects = ProjectFilter.filter_on_color(
            current_projects, Project.Color.from_string(old_color))
        if not current_projects:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no projects exist with this color({old_color})'
                       f' in this workspace({workspace.name}).')
            return

    if old_name:
        current_projects = ProjectFilter.filter_on_name(
            current_projects, old_name)
        if not current_projects:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no projects exist with this name({old_name}'
                       f' in this workspace({workspace.name}).')
            return

    if not multiple and len(current_projects) > 1:
        click.echo(click.style('ERROR', fg='red') +
                   ': multiple projects found matching the criteria, but --multiple '
                   'not specified.. please used the --multiple option or tighten '
                   'the search criteria.')
        return

    current_builds: List[ProjectBuilder] = []
    for project in current_projects:
        current_builds.append(
            ProjectBuilder(project))

    if new_name:
        for build in current_builds:
            assert isinstance(build, ProjectBuilder)
            build.name(new_name)
    if new_color:
        for build in current_builds:
            assert isinstance(build, ProjectBuilder)
            build.color(Project.Color.from_string(new_color))

    cache = retrieve_cache_from_context(context.obj)
    commands = retrieve_commands_from_context(context.obj)

    try:
        built_projects = []
        for project_builder in current_builds:
            project = project_builder.build()
            project = commands.update_project(project)
            click.echo(click.style('SUCCESS', fg='green') +
                       f': updated project({project.name}) in'
                       f' workspace({workspace.name}).')
            built_projects.append(project)
        cache.update_project_cache(built_projects)

    except Exception as e:
        logging.getLogger(__name__).error(e)
        click.echo(click.style('ERROR', fg='red') +
                   f': failed to update the project(s). An exception has been logged;'
                   ' check the logs for more information.')


@projects.command('list')
@click.option('--name')
@click.option('--color',
              type=click.Choice([val.__str__() for val in Project.Color]))
@click.option('--sort-by',
              type=click.Choice([val.__str__() for val in ProjectView.headers()]),
              default=ProjectView.headers()[0],
              show_default=True)
@click.pass_context
def project_list(context: click.Context, name: str, color: str, sort_by: str):
    # Workspaces are already filtered if needed, otherwise this will
    # show all projects in all workspaces.
    workspaces = context.obj['data']['workspaces']

    # Projects are already filtered for any specified workspace.
    current_projects = context.obj['data']['projects']
    if not current_projects:
        click.echo(click.style('WARNING', fg='yellow') +
                   f': no projects found in the specified workspace(s).')
        return

    if name:
        current_projects = ProjectFilter.filter_on_name(
            current_projects, name)
        if not current_projects:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no projects exist with this name({name}).')
            return

    if color:
        # See if the user specified a color to search on.
        current_projects = ProjectFilter.filter_on_color(
            current_projects, Project.Color.from_string(color))
        if not current_projects:
            click.echo(click.style('WARNING', fg='yellow') +
                       f': no projects exist with this color({color}).')
            return

    rows = []
    for current_workspace in workspaces:
        filtered_projects = ProjectFilter.filter_on_workspace(
            current_projects, current_workspace)
        if filtered_projects:
            rows.extend(
                ProjectView(
                    projects=filtered_projects,
                    workspace=current_workspace
                ).values()
            )

    sort_by_index = 0
    if sort_by:
        for index, header in enumerate(ProjectView.headers()):
            if sort_by == header:
                sort_by_index = index
                break

    rows = sorted(rows, key=lambda key: key[sort_by_index], reverse=True)
    click.echo_via_pager(tabulate(
        rows,
        headers=ProjectView.headers(),
        tablefmt="grid"))



from typing import List, Tuple

from togglcmder.toggl.types.tag import Tag as TagType
from togglcmder.toggl.types.workspace import Workspace as WorkspaceType


class Tag(object):
    """
    A composition of the data we want to show in a view
    of a given tag.
    """

    def __init__(self, tags: List[TagType], workspace: WorkspaceType):
        self.__tags = tags
        self.__workspace = workspace

    @staticmethod
    def headers() -> Tuple:
        return (
            "Tag Identifier",
            "Tag Name",
            "Workspace Name"
        )

    def values(self) -> List[Tuple]:
        return [(
            tag.identifier,
            tag.name,
            self.__workspace.name,
        ) for tag in self.__tags]
